import pytest
import re
import uuid

# mypy: disable-error-code="attr-defined"
from tarantool import Datetime as tt_datetime

from conftest import (
    Cluster,
    Instance,
    KeyDef,
    KeyPart,
    Retriable,
    ReturnError,
    TarantoolError,
)


def test_pico_sql(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    usage_msg = re.escape("Usage: sql(query[, params, options])")
    with pytest.raises(ReturnError, match=usage_msg):
        i1.call(
            "pico.sql",
        )
    with pytest.raises(ReturnError, match=usage_msg):
        i1.call(
            "pico.sql",
            "select * from t",
            {},
            False,
            "query_id",
            "extra",
        )

    first_arg_msg = re.escape("SQL query must be a string")
    with pytest.raises(ReturnError, match=first_arg_msg):
        i1.call(
            "pico.sql",
            1,
        )

    second_arg_msg = re.escape("SQL params must be a table")
    with pytest.raises(ReturnError, match=second_arg_msg):
        i1.call(
            "pico.sql",
            "select * from t",
            1,
        )

    third_arg_msg = "SQL options must be a table"
    with pytest.raises(ReturnError, match=third_arg_msg):
        i1.call("pico.sql", "select * from t", {}, "tracer")

    invalid_meta_msg = re.escape("sbroad: table")
    with pytest.raises(ReturnError, match=invalid_meta_msg):
        i1.call(
            "pico.sql",
            "select * from absent_table",
        )
    with pytest.raises(ReturnError, match=invalid_meta_msg):
        i1.call(
            "pico.sql",
            "select * from absent_table where a = ?",
            (1,),
        )


def test_cache_works_for_dml_query(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    def assert_cache_miss(query_id):
        data = i1.eval(
            f"""
            return box.execute([[
            select "span", "query_id" from "_sql_stat"
            where "span" = '"tarantool.cache.miss.read.prepared"' and "query_id" = '{query_id}'
            ]])
        """
        )
        assert len(data["rows"]) == 1

    def assert_cache_hit(query_id):
        data = i1.eval(
            f"""
            return box.execute([[
            select "span", "query_id" from "_sql_stat"
            where "span" = '"tarantool.cache.hit.read.prepared"' and "query_id" = '{query_id}'
            ]])
        """
        )
        assert len(data["rows"]) == 1

    ddl = i1.sql(
        """
        create table t (a int, primary key (a))
        using memtx
        distributed by (a)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    ddl = i1.sql(
        """
        create table not_t (b int, primary key (b))
        using memtx
        distributed by (b)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    data = i1.sql(
        """
        insert into not_t values (1), (2)
    """
    )
    assert data["row_count"] == 2

    query = """
    INSERT INTO t
    SELECT b + 10
    FROM not_t
    ON CONFLICT DO REPLACE
    """
    query_id = "id"

    data = i1.sql(query, options={"traceable": True, "query_id": query_id})
    assert data["row_count"] == 2
    assert_cache_miss(query_id)

    data = i1.sql(query, options={"traceable": True, "query_id": query_id})
    assert data["row_count"] == 2
    assert_cache_hit(query_id)

    # for dml sbroad uses tarantool api,
    # so only dql part of the insert is cached.
    # Check we can reuse it for other query
    id2 = "id2"
    data = i1.sql(
        """
    SELECT b + 10
    FROM not_t
    """,
        options={"traceable": True, "query_id": id2},
    )
    assert data["rows"] == [[11], [12]]
    assert_cache_hit(id2)


def test_tracing(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    query_id = "1"
    exec_cnt = 2
    for i in range(exec_cnt):
        data = i1.sql(
            'select * from "_pico_table"',
            options={"traceable": True, "query_id": query_id},
        )
        assert len(data["rows"]) > 0

    # check we can get the most expensive query using local sql
    data = i1.eval(
        """
        return box.execute([[with recursive st as (
        select * from "_sql_stat" where "query_id" in (select qt."query_id" from qt)
                 and "parent_span" = ''
            union all
            select s.* from "_sql_stat" as s, st on s."parent_span" = st."span"
                 and s."query_id" in (select qt."query_id" from qt)
            ), qt as (
            select s."query_id" from "_sql_stat" as s
            join "_sql_query" as q
                on s."query_id" = q."query_id"
            order by s."count" desc
            limit 1
            )
            select * from st
            where "parent_span" = '';
        ]])
        """
    )
    assert len(data["rows"]) == 1
    query_id_pos = [
        i for i, item in enumerate(data["metadata"]) if item["name"] == "query_id"
    ][0]
    assert data["rows"][0][query_id_pos] == query_id
    exec_cnt_pos = [
        i for i, item in enumerate(data["metadata"]) if item["name"] == "count"
    ][0]
    assert data["rows"][0][exec_cnt_pos] == exec_cnt


def test_select(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table t (a int, primary key (a))
        using memtx
        distributed by (a)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    data = i1.sql("""insert into t values(1);""")
    assert data["row_count"] == 1
    i1.sql("""insert into t values(2);""")
    i1.sql("""insert into t values(?);""", 2000)
    data = i1.sql("""select * from t where a = ?""", 2)
    assert data["rows"] == [[2]]
    data = i1.sql("""select * from t""")
    assert data["rows"] == [[1], [2], [2000]]
    data = i1.sql(
        """select * from t as t1
           join (select a as a2 from t) as t2
           on t1.a = t2.a2 where t1.a = ?""",
        2,
    )
    assert data["rows"] == [[2, 2]]


# test is checking create virtual table with type uuid and cast uuid to text#
def test_uuid(
    cluster: Cluster,
):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    # declaration uuid for test
    t1_id1 = "7055211c-826d-4da2-921b-7133811239f0"
    t1_id2 = "5bc3cc1c-1819-4ab7-adfe-ee0fa2c0cde0"
    t2_id1 = "e4166fc5-e113-46c5-8ae9-970882ca8842"
    t2_id2 = "6f2ba4c4-0a4c-4d79-86ae-43d4f84b70e1"

    # first table with uuid is creating
    ddl = i1.sql(
        """
        create table t1 (id uuid, primary key (id))
        using memtx
        distributed by (id)
        option (timeout = 3)
    """
    )
    # check the creation of the first table
    assert ddl["row_count"] == 1

    # second table has reference by first table
    ddl = i1.sql(
        """
        create table t2 (id uuid not null,t1_id uuid not null, primary key (id))
        using memtx
        distributed by (id)
        option (timeout = 3)
    """
    )
    # check the creation of the second table
    assert ddl["row_count"] == 1

    # table is filling with uuid
    # start
    data = i1.sql("""insert into t1 values(?);""", t1_id1)
    assert data["row_count"] == 1

    i1.sql("""insert into t1 values(?);""", t1_id2)
    assert data["row_count"] == 1

    data = i1.sql("""insert into t2 values(?,?);""", t2_id1, t1_id1)
    assert data["row_count"] == 1

    data = i1.sql("""insert into t2 values(?,?);""", t2_id2, t1_id2)
    assert data["row_count"] == 1
    # end

    # checking virtual table creation
    data = i1.sql(
        """select * from t1 where id in (select t1_id from t2 where id = (?))""",
        uuid.UUID(t2_id1),
    )
    assert data["rows"] == [[uuid.UUID(t1_id1)]]

    # checking cast uuid as text
    data = i1.sql("""select cast(id as Text) from t1""", t1_id1)
    assert data == {
        "metadata": [{"name": "COL_1", "type": "string"}],
        "rows": [[t1_id2], [t1_id1]],
    }


def test_pg_params(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table t (a int, b int, primary key (a))
        using memtx
        distributed by (a)
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    data = i1.sql("""insert into t values ($1, $1), ($2, $1)""", 1, 2)
    assert data["row_count"] == 2

    data = i1.sql("""select * from t""")
    assert data["rows"] == [[1, 1], [2, 1]]

    data = i1.sql(
        """
        select b + $1 from t
        group by b + $1
        having sum(a) > $1
    """,
        1,
    )
    assert data["rows"] == [[2]]

    data = i1.sql(
        """
        select $3, $2, $1, $2, $3 from t
        where a = $1
    """,
        1,
        2,
        3,
    )
    assert data["rows"] == [[3, 2, 1, 2, 3]]

    with pytest.raises(TarantoolError, match="invalid parameters usage"):
        i1.sql(
            """
            select $1, ? from t
            """
        )


def test_read_from_global_tables(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    ddl = i1.sql(
        """
        create table "global_t" ("id" unsigned not null, primary key("id"))
        using memtx
        distributed globally
        """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into "global_t" values (1)
        """
    )
    assert dml["row_count"] == 1

    data = i2.sql(
        """
        select * from "global_t"
        """,
    )
    assert data["rows"] == [[1]]

    data = i1.sql(
        """
        select * from "_pico_table" where "name" = 'global_t'
        """,
    )
    assert len(data["rows"]) == 1


def test_read_from_system_tables(cluster: Cluster):
    instance_count = 2
    cluster.deploy(instance_count=instance_count)
    i1, _ = cluster.instances
    data = i1.sql(
        """
        select * from "_pico_property"
        """,
    )
    assert data["metadata"] == [
        {"name": "key", "type": "string"},
        {"name": "value", "type": "any"},
    ]
    assert len(data["rows"]) == 16

    data = i1.sql(
        """
        select * from "_pico_instance"
        """,
    )
    assert data["metadata"] == [
        {"name": "instance_id", "type": "string"},
        {"name": "instance_uuid", "type": "string"},
        {"name": "raft_id", "type": "unsigned"},
        {"name": "replicaset_id", "type": "string"},
        {"name": "replicaset_uuid", "type": "string"},
        {"name": "current_grade", "type": "array"},
        {"name": "target_grade", "type": "array"},
        {"name": "failure_domain", "type": "map"},
        {"name": "tier", "type": "string"},
    ]
    assert len(data["rows"]) == instance_count


def test_dml_on_global_tbls(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    ddl = i1.sql(
        """
        create table t (x int not null, y int not null, primary key (x))
        using memtx
        distributed by (y)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    ddl = i1.sql(
        """
        create table global_t (id int not null, a int not null, primary key (id))
        using memtx
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    data = i2.sql("insert into t values (1, 1), (2, 2), (3, 3), (4, 4), (5, 5)")
    assert data["row_count"] == 5

    data = i2.sql("insert into global_t values (1, 1), (2, 2)")
    assert data["row_count"] == 2

    # insert into global from global
    data = i2.sql("insert into global_t select id + 2, a from global_t")
    assert data["row_count"] == 2

    # empty insert
    data = i2.sql(
        """
        insert into global_t select id + 2, a from global_t
        where false
        """,
    )
    assert data["row_count"] == 0

    data = i2.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 2], [3, 1], [4, 2]]
    data = i1.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 2], [3, 1], [4, 2]]

    # check update
    data = i2.sql("update global_t set a = 1")
    assert data["row_count"] == 4
    data = i2.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 1], [3, 1], [4, 1]]
    data = i1.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 1], [3, 1], [4, 1]]

    # check update global from global table
    data = i2.sql(
        """
        update global_t
        set a = u from (select a + 1 as u, id as v from global_t) as s
        where id = v
        """
    )
    assert data["row_count"] == 4
    data = i2.sql("select * from global_t")
    assert data["rows"] == [[1, 2], [2, 2], [3, 2], [4, 2]]
    data = i1.sql("select * from global_t")
    assert data["rows"] == [[1, 2], [2, 2], [3, 2], [4, 2]]

    # empty update
    data = i2.sql(
        """
        update global_t
        set a = 20
        where false
        """
    )
    assert data["row_count"] == 0

    # check update global from sharded table
    data = i2.sql(
        """
        update global_t
        set a = y from t
        where id = x
        """
    )
    assert data["row_count"] == 4
    data = i1.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 2], [3, 3], [4, 4]]
    data = i2.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 2], [3, 3], [4, 4]]

    # check delete
    data = i2.sql("delete from global_t")
    assert data["row_count"] == 4

    # test reading subtree with motion
    def q1():
        data = i1.sql("insert into global_t select count(*), 1 from t")
        assert data["row_count"] == 1

    Retriable(rps=5, timeout=10).call(q1)
    data = i1.sql("select * from global_t")
    assert data["rows"] == [[5, 1]]
    data = i2.sql("select * from global_t")
    assert data["rows"] == [[5, 1]]

    # test explain
    lines = i1.sql("explain insert into global_t select * from t")
    expected_explain = """insert "GLOBAL_T" on conflict: fail
    motion [policy: full]
        projection ("T"."X"::integer -> "X", "T"."Y"::integer -> "Y")
            scan "T"
execution options:
sql_vdbe_max_steps = 45000
vtable_max_rows = 5000"""
    assert "\n".join(lines) == expected_explain

    # empty delete
    data = i2.sql("delete from global_t")
    assert data["row_count"] == 1
    data = i2.sql("delete from global_t where false")
    assert data["row_count"] == 0

    data = i2.sql("select * from global_t")
    assert data["rows"] == []
    data = i1.sql("select * from global_t")
    assert data["rows"] == []

    # insert from sharded table
    data = i2.sql("insert into global_t select x, y from t")
    assert data["row_count"] == 5

    data = i2.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
    data = i1.sql("select * from global_t")
    assert data["rows"] == [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]

    def check_table_contents(contents):
        data = i1.sql("select * from t")
        assert sorted(data["rows"]) == contents

    # insert into sharded table from global table
    data = i2.sql("insert into t select id + 5, a + 5 from global_t where id = 1")
    assert data["row_count"] == 1
    Retriable(rps=5, timeout=10).call(
        check_table_contents, [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6]]
    )

    # update sharded table from global table
    data = i2.sql("update t set y = a * a from global_t where id = x")
    assert data["row_count"] == 5
    Retriable(rps=5, timeout=10).call(
        check_table_contents, [[1, 1], [2, 4], [3, 9], [4, 16], [5, 25], [6, 6]]
    )

    # delete sharded table using global table in predicate
    data = i2.sql("delete from t where x in (select id from global_t)")
    assert data["row_count"] == 5
    Retriable(rps=5, timeout=10).call(check_table_contents, [[6, 6]])


def test_datetime(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table t (a int, d datetime not null, primary key (a))
        using memtx
        distributed by (a)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    data = i1.sql(
        """
        insert into t select cast(COLUMN_5 as int), to_date(COLUMN_6, '%Y %d %m') from (values
            (1, '2010 10 10'),
            (2, '2020 20 02'),
            (3, '2010 10 10')
        )
        """
    )
    assert data["row_count"] == 3

    data = i1.sql("""select d from t""")
    assert data["rows"] == [
        [tt_datetime(year=2010, month=10, day=10)],
        [tt_datetime(year=2020, month=2, day=20)],
        [tt_datetime(year=2010, month=10, day=10)],
    ]

    # invalid format
    # FIXME: better error message
    with pytest.raises(TarantoolError, match="could not parse"):
        i1.sql("""select to_date('2020/20/20', '%Y/%d/%m') from t where a = 1""")

    # check we can group on datetime column
    data = i1.sql("""select d from t group by d""")
    assert data["rows"] == [
        [tt_datetime(year=2010, month=10, day=10)],
        [tt_datetime(year=2020, month=2, day=20)],
    ]

    # check we can compare on datetime column
    data = i1.sql("""select d from t where d < to_date('2015/01/01', '%Y/%m/%d')""")
    assert data["rows"] == [
        [tt_datetime(year=2010, month=10, day=10)],
        [tt_datetime(year=2010, month=10, day=10)],
    ]

    data = i1.sql("""select d from t where d = to_date('2020/02/20', '%Y/%m/%d')""")
    assert data["rows"] == [
        [tt_datetime(year=2020, month=2, day=20)],
    ]

    data = i1.sql("""select d from t where d > to_date('2010/12/10', '%Y/%m/%d')""")
    assert data["rows"] == [
        [tt_datetime(year=2020, month=2, day=20)],
    ]

    # without format argument
    # TODO: tarantool does not allow to skip arguments in sql of a stored procedure,
    # but passing '' looks awful, maybe there is a better approach?
    data = i1.sql("""select to_date('1970-01-01T10:10:10 -3', '') from t where a = 1""")
    assert data["rows"] == [[tt_datetime(year=1970, month=1, day=1, tzoffset=-180)]]

    # check we can use arbitrary expressions returning string inside to_date
    data = i1.sql(
        """select to_date(cast('1970-01-01T10:10:10 -3' as string), '' || '')
                  from t where a = 1"""
    )
    assert data["rows"] == [[tt_datetime(year=1970, month=1, day=1, tzoffset=-180)]]

    # check we can create table sharded by datetime column
    ddl = i1.sql(
        """
        create table t2 (a int, d datetime not null, primary key (a))
        using memtx
        distributed by (d)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    # check we can insert min/max date
    data = i1.sql(
        """
        insert into t2 select cast(COLUMN_3 as int), to_date(COLUMN_4, '') from (values
            (1, '-9999-12-31T23:59:59Z'),
            (2, '9999-12-31T23:59:59Z')
        )
        """
    )
    assert data["row_count"] == 2

    # check we can't insert out of limits date
    # FIXME: https://git.picodata.io/picodata/picodata/sbroad/-/issues/639
    with pytest.raises(TarantoolError, match="failed to decode tuple"):
        i1.sql(
            """
            insert into t2 select cast(COLUMN_1 as int), to_date(COLUMN_2, '') from (values
                (1, '-10000-01-01T00:00:00Z')
            )
            """
        )

    with pytest.raises(TarantoolError, match="failed to decode tuple"):
        i1.sql(
            """
            insert into t2 select cast(COLUMN_1 as int), to_date(COLUMN_2, '') from (values
                (2, '10000-01-01T00:00:00Z')
            )
            """
        )
    # test to_char builtin function
    data = i1.sql("""select to_char(d, '%Y-%m-%d') from t""")
    assert sorted(data["rows"], key=lambda e: e[0]) == [
        ["2010-10-10"],
        ["2010-10-10"],
        ["2020-02-20"],
    ]

    # check to_char with timezone specified
    data = i1.sql(
        """select to_char(to_date(COLUMN_1, ''), '%Y-%m-%d')
                  from (values (('1970-01-01T10:10:10 -3')))"""
    )
    assert sorted(data["rows"], key=lambda e: e[0]) == [["1970-01-01"]]

    data = i1.sql(
        """select to_char(to_date(COLUMN_1, ''), '%Y-%Y-%Y')
                  from (values (('1970-01-01T10:10:10 -3')))"""
    )
    assert sorted(data["rows"], key=lambda e: e[0]) == [["1970-1970-1970"]]

    data = i1.sql(
        """select to_char(to_date(COLUMN_1, ''), '%Y-%m-%d %z')
                  from (values (('1970-01-01T10:10:10 -3')))"""
    )
    assert sorted(data["rows"], key=lambda e: e[0]) == [["1970-01-01 -0300"]]


def test_subqueries_on_global_tbls(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table g (a int, b int not null, primary key (a))
        using memtx
        distributed globally
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into g values (1, 1), (2, 2),
        (3, 3), (4, 4), (5, 5)
        """
    )
    assert dml["row_count"] == 5

    ddl = i1.sql(
        """
        create table s (c int, primary key (c))
        using memtx
        distributed by (c)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1
    data = i1.sql("""insert into s values (1), (2), (3), (10);""")
    assert data["row_count"] == 4

    # TODO: remove retries and add more instances when
    # https://git.picodata.io/picodata/picodata/sbroad/-/issues/542
    # is done.
    def check_sq_with_segment_dist():
        data = i1.sql(
            """
            select b from g
            where b in (select c from s where c in (2, 10))
            """,
            timeout=2,
        )
        assert data["rows"] == [[2]]

    Retriable(rps=5, timeout=6).call(check_sq_with_segment_dist)

    def check_sq_with_single_distribution():
        data = i1.sql(
            """
            select b from g
            where b in (select sum(c) from s)
            """,
            timeout=2,
        )
        assert len(data["rows"]) == 0

    Retriable(rps=5, timeout=6).call(check_sq_with_single_distribution)

    def check_sq_with_any_distribution():
        data = i1.sql(
            """
            select b from g
            where b in (select c * 5 from s)
            """,
            timeout=2,
        )
        assert data["rows"] == [[5]]

    Retriable(rps=5, timeout=6).call(check_sq_with_any_distribution)

    def check_sqs_joined_with_or():
        # first subquery selects [1], [2], [3]
        # second subquery must add additional [4] tuple
        data = i1.sql(
            """
            select b from g
            where b in (select c from s) or a in (select count(*) from s)
            """,
            timeout=2,
        )
        assert data["rows"] == [[1], [2], [3], [4]]

    Retriable(rps=5, timeout=6).call(check_sqs_joined_with_or)

    def check_sqs_joined_with_and():
        data = i1.sql(
            """
            select b from g
            where b in (select c from s) and a in (select count(*) from s)
            """,
            timeout=2,
        )
        assert len(data["rows"]) == 0

    Retriable(rps=5, timeout=6).call(check_sqs_joined_with_and)

    def check_sq_in_join_condition_joined_with_and():
        data = i1.sql(
            """
            select c from s inner join
            (select c as c1 from s)
            on c = c1 + 3 and c in (select a from g)
            """,
            timeout=2,
        )
        assert data["rows"] == []

    Retriable(rps=5, timeout=6).call(check_sq_in_join_condition_joined_with_and)

    def check_sq_in_join_condition_joined_with_or():
        # Full join because of 'OR'
        data = i1.sql(
            """
            select min(c) from s inner join
            (select c as c1 from s)
            on c = c1 + 3 or c in (select a from g)
            """,
            timeout=2,
        )
        assert data["rows"] == [[1]]

    Retriable(rps=5, timeout=6).call(check_sq_in_join_condition_joined_with_or)

    def check_bucket_discovery_when_sqs_joined_with_or():
        data = i1.sql(
            """
            select a from g
            where b in (select c from s where c = 1) or
            b in (select c from s where c = 3)
            """,
            timeout=2,
        )
        assert data["rows"] == [[1], [3]]

    Retriable(rps=5, timeout=6).call(check_bucket_discovery_when_sqs_joined_with_or)

    def check_sqs_joined_with_or_and():
        data = i1.sql(
            """
            select a from g
            where b in (select c from s where c = 1) or
            b in (select c from s where c = 3) and
            a < (select sum(c) from s)
            """,
            timeout=2,
        )
        assert data["rows"] == [[1], [3]]

    Retriable(rps=5, timeout=6).call(check_sqs_joined_with_or_and)


def test_aggregates_on_global_tbl(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table g (a int, b int not null, primary key (a))
        using memtx
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into g values (1, 1), (2, 2), (3, 1),
        (4, 1), (5, 2)
        """
    )
    assert dml["row_count"] == 5

    data = i1.sql(
        """
        select count(*), min(b), max(b), min(b) + max(b) from g
        """
    )
    assert data["rows"] == [[5, 1, 2, 3]]

    data = i1.sql(
        """
        select b*b, sum(a + 1) from g
        group by b*b
        """
    )
    assert data["rows"] == [[1, 11], [4, 9]]

    data = i1.sql(
        """
        select b*b, sum(a + 1) from g
        group by b*b
        having count(a) > 2
        """
    )
    assert data["rows"] == [[1, 11]]


def test_join_with_global_tbls(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table g (a int, b int not null, primary key (a))
        using memtx
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into g values (1, 1), (2, 1), (3, 3)
        """
    )
    assert dml["row_count"] == 3

    ddl = i1.sql(
        """
        create table s (c int, primary key (c))
        using memtx
        distributed by (c)
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1
    data = i1.sql("""insert into s values (1), (2), (3), (4), (5);""")
    assert data["row_count"] == 5

    expected_rows = [[1], [1], [3]]

    def check_inner_join_global_vs_segment():
        data = i1.sql(
            """
            select b from g
            join s on g.a = s.c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == expected_rows

    Retriable(rps=5, timeout=6).call(check_inner_join_global_vs_segment)

    def check_inner_join_segment_vs_global():
        data = i1.sql(
            """
            select b from s
            join g on g.a = s.c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == expected_rows

    Retriable(rps=5, timeout=6).call(check_inner_join_segment_vs_global)

    def check_inner_join_segment_vs_global_sq_in_cond():
        data = i1.sql(
            """
            select c from s
            join g on 1 = 1 and
            c in (select a*a from g)
            group by c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[1], [4]]

    Retriable(rps=5, timeout=6).call(check_inner_join_segment_vs_global_sq_in_cond)

    def check_left_join_segment_vs_global_sq_in_cond():
        data = i1.sql(
            """
            select c, cast(sum(a) as int) from s
            left join g on 1 = 1 and
            c in (select a*a from g)
            where c < 4
            group by c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [
            [1, 6],
            [2, None],
            [3, None],
        ]

    Retriable(rps=5, timeout=6).call(check_left_join_segment_vs_global_sq_in_cond)

    def check_left_join_any_vs_global():
        data = i1.sql(
            """
            select c, b from
            (select c*c as c from s)
            left join g on c = b
            where c < 5
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[1, 1], [1, 1], [4, None]]

    Retriable(rps=5, timeout=6).call(check_left_join_any_vs_global)

    def check_inner_join_any_vs_global():
        data = i1.sql(
            """
            select c, b from
            (select c*c as c from s)
            inner join g on c = b
            where c < 5
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[1, 1], [1, 1]]

    Retriable(rps=5, timeout=6).call(check_inner_join_any_vs_global)

    def check_left_join_single_vs_global():
        data = i1.sql(
            """
            select c, a from
            (select count(*) as c from s)
            left join g on c = a + 2
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[5, 3]]

    Retriable(rps=5, timeout=6).call(check_left_join_single_vs_global)

    def check_left_join_global_with_expr_in_proj_vs_segment():
        data = i1.sql(
            """
            select b, c from (select b + 3 as b from g)
            left join s on b = c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[4, 4], [4, 4], [6, None]]

    Retriable(rps=5, timeout=6).call(
        check_left_join_global_with_expr_in_proj_vs_segment
    )

    def check_left_join_global_vs_any_false_condition():
        data = i1.sql(
            """
            select b, c from g
            left join
            (select c*c as c from s where c > 3)
            on b = c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [
            [1, None],
            [1, None],
            [3, None],
        ]

    Retriable(rps=5, timeout=6).call(check_left_join_global_vs_any_false_condition)

    def check_left_join_global_vs_any():
        data = i1.sql(
            """
            select b, c from g
            left join
            (select c*c as c from s where c < 3)
            on b = c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[1, 1], [1, 1], [3, None]]

    Retriable(rps=5, timeout=6).call(check_left_join_global_vs_any)

    def check_left_join_complex_global_child_vs_any():
        data = i1.sql(
            """
            select a, b, c from (
                select a, b from g
                inner join (select a + 2 as u from g)
                on a = u
            )
            left join
            (select c + 1 as c from s where c = 2)
            on b = c
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda e: e[0]) == [[3, 3, 3]]

    Retriable(rps=5, timeout=6).call(check_left_join_complex_global_child_vs_any)


def test_union_all_on_global_tbls(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table g (a int, b int not null, primary key (a))
        using memtx
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into g values (1, 1), (2, 2), (3, 2)
        """
    )
    assert dml["row_count"] == 3

    ddl = i1.sql(
        """
        create table s (c int, d int not null, primary key (c))
        using memtx
        distributed by (c)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1
    data = i1.sql("""insert into s values (1, 2), (2, 2), (3, 2);""")
    assert data["row_count"] == 3

    expected = [[1], [2], [2], [2], [2], [2]]

    def check_global_vs_any():
        data = i1.sql(
            """
            select b from g
            union all
            select d from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_global_vs_any)

    def check_any_vs_global():
        data = i1.sql(
            """
            select d from s
            union all
            select b from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_any_vs_global)

    def check_global_vs_global():
        data = i1.sql(
            """
            select b from g
            union all
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [
            [1],
            [1],
            [2],
            [2],
            [2],
            [3],
        ]

    Retriable(rps=5, timeout=6).call(check_global_vs_global)

    expected = [[1], [1], [2], [2], [3], [3]]

    def check_global_vs_segment():
        data = i1.sql(
            """
            select a from g
            union all
            select c from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_global_vs_segment)

    def check_segment_vs_global():
        data = i1.sql(
            """
            select c from s
            union all
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_segment_vs_global)

    expected = [[1], [2], [3], [6]]

    def check_single_vs_global():
        data = i1.sql(
            """
            select sum(c) from s
            union all
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_single_vs_global)

    def check_global_vs_single():
        data = i1.sql(
            """
            select a from g
            union all
            select sum(c) from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_global_vs_single)

    # some arbitrary queries

    def check_multiple_union_all():
        data = i1.sql(
            """
            select a from g
            where a = 2
            union all
            select d from s
            group by d
            union all
            select a from g
            where b = 1
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[1], [2], [2]]

    Retriable(rps=5, timeout=6).call(check_multiple_union_all)

    def check_union_with_where():
        data = i1.sql(
            """
            select a from g
            where a in (select d from s)
            union all
            select c from s
            where c = 3
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[2], [3]]

    Retriable(rps=5, timeout=6).call(check_union_with_where)

    def check_complex_segment_child():
        data = i1.sql(
            """
            select a, b from g
            where a in (select d from s)
            union all
            select d, sum(u) from s
            inner join (select c as u from s)
            on d = u or u = 1
            group by d
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[2, 2], [2, 9]]

    Retriable(rps=5, timeout=6).call(check_complex_segment_child)


def test_union_on_global_tbls(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table g (a int, b int not null, primary key (a))
        using memtx
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    for i, j in [(1, 1), (2, 2), (3, 2)]:
        index = i1.cas("insert", "G", [i, j])
        i1.raft_wait_index(index, 3)

    ddl = i1.sql(
        """
        create table s (c int, d int not null, primary key (c))
        using memtx
        distributed by (c)
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1
    data = i1.sql("""insert into s values (1, 2), (2, 2), (3, 2);""")
    assert data["row_count"] == 3

    expected = [[1], [2]]

    def check_global_vs_any():
        data = i1.sql(
            """
            select b from g
            union
            select d from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_global_vs_any)

    def check_any_vs_global():
        data = i1.sql(
            """
            select d from s
            union
            select b from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_any_vs_global)

    def check_global_vs_global():
        data = i1.sql(
            """
            select b from g
            union
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [
            [1],
            [2],
            [3],
        ]

    Retriable(rps=5, timeout=6).call(check_global_vs_global)

    expected = [[1], [2], [3]]

    def check_global_vs_segment():
        data = i1.sql(
            """
            select a from g
            union
            select c from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_global_vs_segment)

    def check_segment_vs_global():
        data = i1.sql(
            """
            select c from s
            union
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_segment_vs_global)

    expected = [[1], [2], [3]]

    def check_single_vs_global():
        data = i1.sql(
            """
            select sum(c) - 3 from s
            union
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_single_vs_global)

    def check_global_vs_single():
        data = i1.sql(
            """
            select a from g
            union
            select sum(c) - 3 from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == expected

    Retriable(rps=5, timeout=6).call(check_global_vs_single)

    def check_multiple_union():
        data = i1.sql(
            """
            select a from g
            where a = 2
            union
            select d from s
            group by d
            union
            select a from g
            where b = 1
            except
            select null from g
            where false
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[1], [2]]

    Retriable(rps=5, timeout=6).call(check_multiple_union)


def test_trim(instance: Instance):
    instance.sql(
        """
        create table t (s string, primary key (s))
        using memtx
        distributed by (s)
        """
    )

    instance.sql(""" insert into t values (' aabb ') """)

    # basic trim test
    data = instance.sql(""" select trim(s) from t """)
    assert data["rows"][0] == ["aabb"]

    # trim inside trim
    data = instance.sql(""" select trim(trim(s)) from t """)
    assert data["rows"][0] == ["aabb"]

    # trim with removal chars
    data = instance.sql(""" select trim('a' from trim(s)) from t """)
    assert data["rows"][0] == ["bb"]

    data = instance.sql(""" select trim(trim(s) from trim(s)) from t """)
    assert data["rows"][0] == [""]

    # trim with modifiers
    data = instance.sql(""" select trim(leading 'a' from trim(s)) from t """)
    assert data["rows"][0] == ["bb"]

    data = instance.sql(""" select trim(trailing 'b' from trim(s)) from t """)
    assert data["rows"][0] == ["aa"]

    data = instance.sql(""" select trim(both 'ab' from trim(s)) from t """)
    assert data["rows"][0] == [""]


def test_except_on_global_tbls(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table g (a int, b int not null, primary key (a))
        using memtx
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into g values (1, 1), (2, 2),
        (3, 3), (4, 4), (5, 5)
        """
    )
    assert dml["row_count"] == 5

    ddl = i1.sql(
        """
        create table s (c int, d int not null, primary key (c))
        using memtx
        distributed by (c)
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    data = i1.sql("""insert into s values (3, 2), (4, 3), (5, 4), (6, 5), (7, 6);""")
    assert data["row_count"] == 5

    def check_global_vs_global():
        data = i1.sql(
            """
            select a from g
            except
            select a - 1 from g
            """,
            timeout=2,
        )
        assert data["rows"] == [[5]]

    Retriable(rps=5, timeout=6).call(check_global_vs_global)

    def check_global_vs_segment():
        data = i1.sql(
            """
            select a from g
            except
            select c from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[1], [2]]

    Retriable(rps=5, timeout=6).call(check_global_vs_segment)

    def check_global_vs_any():
        data = i1.sql(
            """
            select b from g
            except
            select d from s
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[1]]

    Retriable(rps=5, timeout=6).call(check_global_vs_any)

    def check_global_vs_single():
        data = i1.sql(
            """
            select b from g
            where b = 1 or b = 2
            except
            select sum(d) from s
            where d = 3
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[1], [2]]

    Retriable(rps=5, timeout=6).call(check_global_vs_single)

    def check_single_vs_global():
        data = i1.sql(
            """
            select sum(d) from s
            where d = 3 or d = 2
            except
            select b from g
            where b = 1 or b = 2
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[5]]

    Retriable(rps=5, timeout=6).call(check_single_vs_global)

    def check_segment_vs_global():
        data = i1.sql(
            """
            select c from s
            except
            select a from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[6], [7]]

    Retriable(rps=5, timeout=6).call(check_segment_vs_global)

    def check_any_vs_global():
        data = i1.sql(
            """
            select d from s
            except
            select b from g
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[6]]

    Retriable(rps=5, timeout=6).call(check_any_vs_global)

    def check_multiple_excepts():
        data = i1.sql(
            """
            select a + 5 from g
            where a = 1 or a = 2
            except select * from (
            select d from s
            except
            select b from g
            )
            """,
            timeout=2,
        )
        assert sorted(data["rows"], key=lambda x: x[0]) == [[7]]

    Retriable(rps=5, timeout=6).call(check_multiple_excepts)


def test_hash(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table t (a int, primary key (a))
        using memtx
        distributed by (a)
    """
    )
    assert ddl["row_count"] == 1

    # Calculate tuple hash with Lua
    tup = (1,)
    key_def = KeyDef([KeyPart(1, "integer", True)])
    lua_hash = i1.hash(tup, key_def)
    bucket_count = 3000

    # Compare SQL and Lua bucket_id
    data = i1.sql("""insert into t values(?);""", 1)
    assert data["row_count"] == 1
    data = i1.sql(""" select "bucket_id" from t where a = ?""", 1)
    assert data["rows"] == [[lua_hash % bucket_count + 1]]


def test_select_lowercase_name(cluster: Cluster):
    i1, *_ = cluster.deploy(instance_count=1)

    ddl = i1.sql(
        """
        create table "lowercase_name" ("id" int, primary key ("id"))
        distributed by ("id")
    """
    )
    assert ddl["row_count"] == 1

    assert i1.call("box.space.lowercase_name:select") == []

    data = i1.sql("""insert into "lowercase_name" values(420);""")
    assert data["row_count"] == 1
    data = i1.sql("""select * from "lowercase_name" """)
    assert data["rows"] == [[420]]


def test_select_string_field(cluster: Cluster):
    i1, *_ = cluster.deploy(instance_count=1)

    ddl = i1.sql(
        """
        create table "STUFF" ("id" integer not null, "str" string null, primary key ("id"))
        distributed by ("id")
    """
    )
    assert ddl["row_count"] == 1

    data = i1.sql("""insert into STUFF values(1337, 'foo');""")
    assert data["row_count"] == 1
    data = i1.sql("""select * from STUFF """)
    assert data["rows"] == [[1337, "foo"]]


def test_create_drop_table(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    ddl = i1.sql(
        """
        create table "t" ("a" integer, "b" int, primary key ("b", "a"))
        using memtx
        distributed by ("a", "b")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    # Already exists -> ok.
    ddl = i1.sql(
        """
        create table "t" ("a" integer, "b" int, primary key ("b", "a"))
        using memtx
        distributed by ("a", "b")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 0

    # FIXME: this should fail
    # see https://git.picodata.io/picodata/picodata/picodata/-/issues/331
    # Already exists with different format -> error.
    ddl = i1.sql(
        """
        create table "t" ("key" string, "value" string not null, primary key ("key"))
        using memtx
        distributed by ("key")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 0

    ddl = i2.sql(
        """
        drop table "t"
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    # already dropped -> error, no such table
    with pytest.raises(TarantoolError, match="""sbroad: space t not found"""):
        i2.sql(
            """
            drop table "t"
            option (timeout = 3)
        """
        )

    ddl = i2.sql(
        """
        create table "t" ("a" integer, "b" int, primary key ("b", "a"))
        using memtx
        distributed by ("a", "b")
    """
    )
    assert ddl["row_count"] == 1

    ddl = i1.sql(
        """
        drop table "t"
    """
    )
    assert ddl["row_count"] == 1

    # Check vinyl space
    ddl = i1.sql(
        """
        create table "t" ("key" string, "value" string not null, primary key ("key"))
        using vinyl
        distributed by ("key")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    ddl = i2.sql(
        """
        drop table "t"
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    # Check NOT NULL inferred on PRIMARY KEY
    ddl = i1.sql(
        """
        create table "t" (a int primary key)
        distributed by (a)
        """
    )
    assert ddl["row_count"] == 1
    with pytest.raises(
        TarantoolError,
        match=(
            "Tuple field 1 \\(A\\) type does not match one required"
            " by operation: expected integer, got nil"
        ),
    ):
        i1.sql('insert into "t" values (null)')

    ddl = i2.sql(
        """
        drop table "t"
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    # Check global space
    ddl = i1.sql(
        """
        create table "global_t" ("key" string, "value" string not null,
        primary key ("key"))
        using memtx
        distributed globally
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    with pytest.raises(TarantoolError, match="global spaces can use only memtx engine"):
        i1.sql(
            """
            create table "t" ("key" string, "value" string not null,
            primary key ("key"))
            using vinyl
            distributed globally
            option (timeout = 3)
            """
        )

    # Check table creation with different forms of primary key declaration.
    with pytest.raises(TarantoolError, match="Primary key has been already declared"):
        i1.sql(
            """
            create table "primary_t" (a int not null primary key, b int not null primary key)
            distributed by (a)
            """
        )
    with pytest.raises(TarantoolError, match="Primary key has been already declared"):
        i1.sql(
            """
            create table "primary_t" (a int primary key, b int, primary key (a))
            distributed by (a)
            """
        )
    with pytest.raises(TarantoolError, match="Primary key has been already declared"):
        i1.sql(
            """
            create table "primary_t" (a int not null primary key, b int, primary key (b))
            distributed by (a)
            """
        )
    with pytest.raises(TarantoolError, match="Primary key must be declared"):
        i1.sql(
            """
            create table "primary_t" (a int)
            distributed by (a)
            """
        )

    ddl = i1.sql(
        """
        create table "primary_t" (a int not null primary key)
        distributed by (a)
        """
    )
    assert ddl["row_count"] == 1


def test_check_format(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    # Primary key missing.
    with pytest.raises(TarantoolError, match="Primary key column b not found"):
        i1.sql(
            """
        create table "error" ("a" integer, primary key ("b"))
        using memtx
        distributed by ("a")
    """
        )
    # Sharding key missing.
    with pytest.raises(TarantoolError, match="Sharding key column b not found"):
        i1.sql(
            """
        create table "error" ("a" integer not null, primary key ("a"))
        using memtx
        distributed by ("b")
    """
        )
    # Nullable primary key.
    with pytest.raises(
        TarantoolError, match="Primary key mustn't contain nullable columns"
    ):
        i1.sql(
            """
        create table "error" ("a" integer null, primary key ("a"))
        using memtx
        distributed by ("a")
    """
        )

    # Check format
    ddl = i1.sql(
        """
        create table "t" (
            "non_nullable" string not null,
            "nullable" Boolean null,
            "default" Decimal,
            primary key ("non_nullable")
            )
        using memtx
        distributed by ("non_nullable")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1
    format = i1.call("box.space.t:format")
    assert format == [
        {"is_nullable": False, "name": "non_nullable", "type": "string"},
        {"is_nullable": False, "name": "bucket_id", "type": "unsigned"},
        {"is_nullable": True, "name": "nullable", "type": "boolean"},
        {"is_nullable": True, "name": "default", "type": "decimal"},
    ]

    # Inserting with nulls/nonnulls works.
    dml = i1.sql(
        """
        insert into "t" values
            ('Name1', true, 0.0),
            ('Name2', null, -0.12974679036997294),
            ('Name3', false, null)
    """
    )
    assert dml["row_count"] == 3
    # Inserting with nulls/nonnulls works using params.
    dml = i1.sql(
        """
        insert into "t" ("non_nullable", "nullable") values
            ('Name4', ?),
            ('Name5', ?)
    """,
        True,
        None,
    )
    assert dml["row_count"] == 2


def test_insert(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, _ = cluster.instances

    ddl = i1.sql(
        """
        create table "t" ("a" integer not null, primary key ("a"))
        using memtx
        distributed by ("a")
    """
    )
    assert ddl["row_count"] == 1

    # Wrong parameters number.
    with pytest.raises(
        TarantoolError,
        match="invalid node: parameter node does not refer to an expression",
    ):
        i1.sql(
            """
        insert into "t" values (?)
        """
        )
    with pytest.raises(
        TarantoolError, match="Expected at least 2 values for parameters. Got 1"
    ):
        i1.sql(
            """
        insert into "t" values (?), (?)
            """,
            1,
        )


def test_insert_on_conflict(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table "t" ("a" integer not null, "b" int not null, primary key ("a"))
        using memtx
        distributed by ("b")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into "t" values (1, 1)
    """
    )
    assert dml["row_count"] == 1

    dml = i1.sql(
        """
        insert into "t" values (1, 1) on conflict do nothing
    """
    )
    assert dml["row_count"] == 0

    data = i1.sql(
        """select * from "t"
    """
    )
    assert data["rows"] == [[1, 1]]

    dml = i1.sql(
        """
        insert into "t" values (1, 2) on conflict do replace
    """
    )
    assert dml["row_count"] == 1

    data = i1.sql(
        """select * from "t"
    """
    )
    assert data["rows"] == [[1, 2]]


def test_sql_limits(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table "t" ("a" integer not null, "b" int not null, primary key ("a"))
        using memtx
        distributed by ("b")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
    insert into "t" values (1, 1), (2, 1)
    """
    )
    assert dml["row_count"] == 2

    with pytest.raises(
        TarantoolError, match="Reached a limit on max executed vdbe opcodes. Limit: 5"
    ):
        i1.sql(
            """
        select * from "t" where "a" = 1 option(sql_vdbe_max_steps=5)
    """
        )

    dql = i1.sql(
        """
        select * from "t" where "a" = 1 option(sql_vdbe_max_steps=50)
    """
    )
    assert dql["rows"] == [[1, 1]]

    with pytest.raises(
        TarantoolError,
        match=r"Exceeded maximum number of rows \(1\) in virtual table: 2",
    ):
        i1.sql(
            """
        select * from "t" option(vtable_max_rows=1, sql_vdbe_max_steps=50)
    """
        )


def test_sql_acl_password_length(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    username = "USER"
    password_short = "pwd"
    password_long = "Passw0rd"

    acl = i1.sql(
        f"""
    create user {username} with password '{password_long}'
    using md5 option (timeout = 3)
"""
    )
    assert acl["row_count"] == 1
    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1

    with pytest.raises(TarantoolError, match="password is too short"):
        i1.sql(
            """
            create user {username} with password '{password}'
            using md5 option (timeout = 3)
        """.format(
                username=username, password=password_short
            )
        )

    acl = i1.sql(
        f"""
    create user {username} with password '{password_short}'
    using ldap option (timeout = 3)
"""
    )
    assert acl["row_count"] == 1
    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1


def test_sql_acl_users_roles(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    username = "User"
    password = "Passw0rd"
    upper_username = "USER"
    rolename = "Role"
    upper_rolename = "ROLE"
    acl = i1.sql(
        f"""
        create user "{username}" with password '{password}'
        using md5 option (timeout = 3)
    """
    )
    assert acl["row_count"] == 1
    assert (
        i1.call("box.space._pico_user.index._pico_user_name:get", username) is not None
    )

    # Dropping user that doesn't exist should return 0.
    acl = i1.sql(f"drop user {upper_username}")
    assert acl["row_count"] == 0

    # Dropping user that does exist should return 1.
    acl = i1.sql(f'drop user "{username}"')
    assert acl["row_count"] == 1
    assert i1.call("box.space._pico_user.index._pico_user_name:get", username) is None

    # All the usernames below should match the same user.
    # * Upcasted username in double parentheses shouldn't change.
    acl = i1.sql(
        f"""
        create user "{upper_username}" password '{password}'
        using chap-sha1
    """
    )
    assert acl["row_count"] == 1
    # * Username as is in double parentheses.
    acl = i1.sql(f'drop user "{upper_username}"')
    assert acl["row_count"] == 1
    acl = i1.sql(
        f"""
        create user "{upper_username}" password '' using ldap
    """
    )
    assert acl["row_count"] == 1
    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1
    # * Username without parentheses should be upcasted.
    acl = i1.sql(
        f"""
        create user {username} with password '{password}'
        option (timeout = 3)
    """
    )
    assert acl["row_count"] == 1
    acl = i1.sql(f'drop user "{upper_username}"')
    assert acl["row_count"] == 1

    # Check user creation with LDAP works well with non-empty password specification
    # (it must be ignored).
    acl = i1.sql(
        f"""
        create user "{upper_username}" password 'smth' using ldap
    """
    )
    assert acl["row_count"] == 1

    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1

    # We can safely retry creating the same user.
    acl = i1.sql(f"create user {username} with password '{password}' using md5")
    assert acl["row_count"] == 1
    acl = i1.sql(f"create user {username} with password '{password}' using md5")
    assert acl["row_count"] == 0
    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1

    # Zero timeout should return timeout error.
    with pytest.raises(TarantoolError, match="timeout"):
        i1.sql(f"drop user {username} option (timeout = 0)")
    with pytest.raises(TarantoolError, match="timeout"):
        i1.sql(f"drop role {username} option (timeout = 0)")
    with pytest.raises(TarantoolError, match="timeout"):
        i1.sql(
            f"""
            create user {username} with password '{password}'
            option (timeout = 0)
        """
        )
    with pytest.raises(TarantoolError, match="timeout"):
        i1.sql(
            f"""
            alter user {username} with password '{password}'
            option (timeout = 0)
        """
        )

    # Username in single quotes is unsupported.
    with pytest.raises(TarantoolError, match="rule parsing error"):
        i1.sql(f"drop user '{username}'")
    with pytest.raises(TarantoolError, match="rule parsing error"):
        i1.sql(f"create user '{username}' with password '{password}'")
    with pytest.raises(TarantoolError, match="rule parsing error"):
        i1.sql(f"alter user '{username}' with password '{password}'")
    # Rolename in single quotes is unsupported.
    with pytest.raises(TarantoolError, match="rule parsing error"):
        i1.sql(f"drop role '{username}'")

    # Can't create same user with different auth methods.
    with pytest.raises(
        TarantoolError, match="already exists with different auth method"
    ):
        i1.sql(f"create user {username} with password '{password}' using md5")
        i1.sql(f"create user {username} with password '{password}' using chap-sha1")

    # Can't create same user with different password.
    with pytest.raises(
        TarantoolError, match="already exists with different auth method"
    ):
        i1.sql(f"create user {username} with password 'Badpa5SS' using md5")
        i1.sql(f"create user {username} with password 'Badpa5SS' using md5")
    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1

    # Attempt to drop role with name of user should return error.
    with pytest.raises(
        TarantoolError, match=f"User {username} exists. Unable to drop user."
    ):
        i1.sql(f""" create user "{username}" with password '{password}' """)
        i1.sql(f""" drop role "{username}" """)
    acl = i1.sql(f""" drop user "{username}" """)
    assert acl["row_count"] == 1

    another_password = "Qwerty123"
    # Alter of unexisted user should do nothing.
    acl = i1.sql(f"alter user \"nobody\" with password '{another_password}'")
    assert acl["row_count"] == 0

    # Check altering works.
    acl = i1.sql(f"create user {username} with password '{password}' using md5")
    assert acl["row_count"] == 1
    user_def = i1.call("box.space._pico_user.index._pico_user_name:get", upper_username)
    users_auth_was = user_def[3]
    # * Password and method aren't changed -> update nothing.
    acl = i1.sql(f"alter user {username} with password '{password}' using md5")
    assert acl["row_count"] == 0
    user_def = i1.call("box.space._pico_user.index._pico_user_name:get", upper_username)
    users_auth_became = user_def[3]
    assert users_auth_was == users_auth_became

    # * Password is changed -> update hash.
    acl = i1.sql(f"alter user {username} with password '{another_password}' using md5")
    assert acl["row_count"] == 1
    user_def = i1.call("box.space._pico_user.index._pico_user_name:get", upper_username)
    users_auth_became = user_def[3]
    assert users_auth_was[0] == users_auth_became[0]
    assert users_auth_was[1] != users_auth_became[1]

    # * Password and method are changed -> update method and hash.
    acl = i1.sql(
        f"alter user {username} with password '{another_password}' using chap-sha1"
    )
    assert acl["row_count"] == 1
    user_def = i1.call("box.space._pico_user.index._pico_user_name:get", upper_username)
    users_auth_became = user_def[3]
    assert users_auth_was[0] != users_auth_became[0]
    assert users_auth_was[1] != users_auth_became[1]
    # * LDAP should ignore password -> update method and hash.
    acl = i1.sql(f"alter user {username} with password '{another_password}' using ldap")
    assert acl["row_count"] == 1
    user_def = i1.call("box.space._pico_user.index._pico_user_name:get", upper_username)
    users_auth_became = user_def[3]
    assert users_auth_was[0] != users_auth_became[0]
    assert users_auth_became[1] == ""
    acl = i1.sql(f"drop user {username}")
    assert acl["row_count"] == 1

    # Attempt to create role with the name of already existed user
    # should lead to an error.
    acl = i1.sql(
        f""" create user "{username}" with password 'Validpassw0rd' using md5 """
    )
    assert acl["row_count"] == 1
    with pytest.raises(TarantoolError, match="User with the same name already exists"):
        i1.sql(f'create role "{username}"')
    acl = i1.sql(f'drop user "{username}"')
    assert acl["row_count"] == 1

    # Dropping role that doesn't exist should return 0.
    acl = i1.sql(f"drop role {rolename}")
    assert acl["row_count"] == 0

    # Successive creation of role.
    acl = i1.sql(f'create role "{rolename}"')
    assert acl["row_count"] == 1
    # Unable to alter role.
    with pytest.raises(
        TarantoolError, match=f"Role {rolename} exists. Unable to alter role."
    ):
        i1.sql(f"alter user \"{rolename}\" with password '{password}'")

    # Attempt to drop user with name of role should return error.
    with pytest.raises(
        TarantoolError, match=f"Role {rolename} exists. Unable to drop role."
    ):
        i1.sql(f""" drop user "{rolename}" """)

    # Creation of the role that already exists shouldn't do anything.
    acl = i1.sql(f'create role "{rolename}"')
    assert acl["row_count"] == 0
    assert (
        i1.call("box.space._pico_user.index._pico_user_name:get", rolename) is not None
    )

    # Dropping role that does exist should return 1.
    acl = i1.sql(f'drop role "{rolename}"')
    assert acl["row_count"] == 1
    assert i1.call("box.space._pico_user.index._pico_user_name:get", rolename) is None

    # All the rolenames below should match the same role.
    acl = i1.sql(f'create role "{upper_rolename}"')
    assert acl["row_count"] == 1
    acl = i1.sql(f"drop role {upper_rolename}")
    assert acl["row_count"] == 1
    acl = i1.sql(f'create role "{upper_rolename}"')
    assert acl["row_count"] == 1
    acl = i1.sql(f"drop role {rolename}")
    assert acl["row_count"] == 1
    acl = i1.sql(f'create role "{upper_rolename}"')
    assert acl["row_count"] == 1
    acl = i1.sql(f'drop role "{upper_rolename}"')
    assert acl["row_count"] == 1


def test_sql_alter_login(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    # Create the owner of `USER`
    owner_username = "OWNER_USER"
    owner_password = "PA5sWORD"

    acl = i1.sudo_sql(f"create user {owner_username} with password '{owner_password}'")
    assert acl["row_count"] == 1

    acl = i1.sudo_sql(f"grant create user to {owner_username}")
    assert acl["row_count"] == 1

    username = "USER"
    password = "PA5sWORD"
    # Create user.
    acl = i1.sql(
        f"create user {username} with password '{password}'",
        user=owner_username,
        password=owner_password,
    )
    assert acl["row_count"] == 1

    # Alter user with LOGIN option - opertaion is idempotent.
    acl = i1.sudo_sql(f"alter user {username} with login")
    assert acl["row_count"] == 1
    # Alter user with NOLOGIN option.
    acl = i1.sudo_sql(f"alter user {username} with nologin")
    assert acl["row_count"] == 1
    # Alter user with NOLOGIN again - operation is idempotent.
    acl = i1.sudo_sql(f"alter user {username} with nologin")
    assert acl["row_count"] == 1
    # Login privilege is removed
    with pytest.raises(
        Exception,
        match="User does not have login privilege",
    ):
        i1.sql("insert into t values(2);", user=username, password=password)

    # Alter user with LOGIN option again - `USER` owner can also modify LOGIN privilege.
    acl = i1.sql(
        f"alter user {username} with login",
        user=owner_username,
        password=owner_password,
    )
    assert acl["row_count"] == 1

    # Alter user with NOLOGIN option again - `USER` owner can also modify LOGIN privilege.
    acl = i1.sql(
        f"alter user {username} with nologin",
        user=owner_username,
        password=owner_password,
    )
    assert acl["row_count"] == 1

    # Login privilege cannot be granted or removed by any user
    # other than admin or user account owner
    other_username = "OTHER_USER"
    other_password = "PA5sWORD"
    acl = i1.sudo_sql(f"create user {other_username} with password '{other_password}'")
    assert acl["row_count"] == 1
    with pytest.raises(
        Exception,
        match=f"Grant Login from '{username}' is denied for {other_username}",
    ):
        acl = i1.sql(
            f"alter user {username} with login",
            user=other_username,
            password=other_password,
        )
    with pytest.raises(
        Exception,
        match=f"Revoke Login from '{username}' is denied for {other_username}",
    ):
        acl = i1.sql(
            f"alter user {username} with nologin",
            user=other_username,
            password=other_password,
        )

    # Login privilege cannot be removed from pico_service even by admin.
    with pytest.raises(
        Exception,
        match="Revoke 'login' from 'pico_service' is denied for all users",
    ):
        i1.sudo_sql('alter user "pico_service" with nologin')


def test_sql_acl_privileges(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    username = "USER"
    another_username = "ANOTHER_USER"
    password = "PaSSW0RD"
    rolename = "ROLE"
    another_rolename = "ANOTHER_ROLE"

    # Create users.
    acl = i1.sql(f"create user {username} with password '{password}'")
    assert acl["row_count"] == 1
    acl = i1.sql(f"create user {another_username} with password '{password}'")
    assert acl["row_count"] == 1
    # Create roles.
    acl = i1.sql(f"create role {rolename}")
    assert acl["row_count"] == 1
    acl = i1.sql(f"create role {another_rolename}")
    assert acl["row_count"] == 1
    # Create tables.
    table_name = "T1"
    another_table_name = "T2"
    ddl = i1.sql(
        f"""
        create table {table_name} ("a" int not null, primary key ("a"))
        distributed by ("a")
    """
    )
    assert ddl["row_count"] == 1

    ddl = i1.sql(
        f"""
        create table {another_table_name} ("a" int not null, primary key ("a"))
        distributed by ("a")
    """
    )
    assert ddl["row_count"] == 1

    # Remember number of default privileges.
    default_privileges_number = len(
        i1.sql(""" select * from "_pico_privilege" """)["rows"]
    )

    # =========================ERRORs======================
    # Attempt to grant unsupported privileges.
    with pytest.raises(
        TarantoolError, match=r"Supported privileges are: \[Read, Write, Alter, Drop\]"
    ):
        i1.sql(f""" grant create on table {table_name} to {username} """)
    with pytest.raises(
        TarantoolError, match=r"Supported privileges are: \[Create, Alter, Drop\]"
    ):
        i1.sql(f""" grant read user to {username} """)
    with pytest.raises(
        TarantoolError, match=r"Supported privileges are: \[Alter, Drop\]"
    ):
        i1.sql(f""" grant create on user {username} to {rolename} """)
    with pytest.raises(
        TarantoolError, match=r"Supported privileges are: \[Create, Drop\]"
    ):
        i1.sql(f""" grant alter role to {username} """)
    with pytest.raises(TarantoolError, match=r"Supported privileges are: \[Drop\]"):
        i1.sql(f""" grant create on role {rolename} to {username} """)

    # Attempt to grant unexisted role.
    with pytest.raises(TarantoolError, match="There is no role with name SUPER"):
        i1.sql(f""" grant SUPER to {username} """)
    # Attempt to grant TO unexisted role.
    with pytest.raises(
        TarantoolError, match="Nor user, neither role with name SUPER exists"
    ):
        i1.sql(f""" grant {rolename} to SUPER """)
    # Attempt to revoke unexisted role.
    with pytest.raises(TarantoolError, match="There is no role with name SUPER"):
        i1.sql(f""" revoke SUPER from {username} """)
    # Attempt to revoke privilege that hasn't been granted yet do noting.
    acl = i1.sql(f""" revoke read on table {table_name} from {username} """)
    assert acl["row_count"] == 0
    # TODO: Attempt to grant role that doesn't visible for user.
    # TODO: Attempt to revoke role that doesn't visible for user.
    # TODO: Attempt to grant TO a user that doesn't visible for user.
    # TODO: Attempt to grant TO a role that doesn't visible for user.
    # TODO: Attempt to revoke FROM a user that doesn't visible for user.
    # TODO: Attempt to revoke FROM a role that doesn't visible for user.

    # ===============ALTER Login/NoLogin=====================
    # TODO: replace with logic from `test_sql_alter_login`.

    # TODO: ================USERs interaction================
    # * TODO: User creation is prohibited.
    # * Grant CREATE to user.
    # Need to grant as admin because only admin is allowed to make wildcard grants
    acl = i1.sudo_sql(f""" grant create user to {username} """)
    assert acl["row_count"] == 1
    # * Check privileges table is updated.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number + 1
    # * TODO: User creation is available.
    # * Revoke CREATE from user.
    acl = i1.sudo_sql(f""" revoke create user from {username} """)
    assert acl["row_count"] == 1
    # * Check privileges table is updated.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number
    # * TODO: Check that user with granted privileges can ALTER and DROP created user
    #         as it's the owner.
    # * TODO: Revoke automatically granted privileges.
    # * TODO: Check ALTER and DROP are prohibited.
    # * Grant global ALTER on users.
    acl = i1.sudo_sql(f""" grant alter user to {username} """)
    assert acl["row_count"] == 1
    # * Check privileges table is updated.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number + 1
    # * TODO: Check ALTER is available.
    # * Revoke global ALTER.
    acl = i1.sudo_sql(f""" revoke alter user from {username} """)
    assert acl["row_count"] == 1

    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number

    # * TODO: Check another user can't initially interact with previously created new user.
    # * TODO: Grant ALTER and DROP user privileges to another user.
    # * TODO: Check user alternation is available.
    # * TODO: Check user drop is available.

    # TODO: ================ROLEs interaction================
    # * TODO: Role creation is prohibited.
    # * Grant CREATE to user.
    acl = i1.sudo_sql(f""" grant create role to {username} """)
    assert acl["row_count"] == 1
    # * Check privileges table is updated.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number + 1
    # * TODO: Role creation is available.
    # * Revoke CREATE from user.
    acl = i1.sudo_sql(f""" revoke create role from {username} """)
    assert acl["row_count"] == 1
    # * Check privileges table is updated.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number
    # * TODO: Check that user with granted privileges can DROP created role as it's the owner.
    # * TODO: Revoke automatically granted privileges.
    # * TODO: Check DROP are prohibited.
    # * Grant global drop on role.
    acl = i1.sudo_sql(f""" grant drop role to {username} """)
    assert acl["row_count"] == 1
    # * Check privileges table is updated.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number + 1
    # * TODO: Check DROP is available.
    # * Revoke global DROP.
    acl = i1.sudo_sql(f""" revoke drop role from {username} """)
    assert acl["row_count"] == 1

    # * TODO: Check another user can't initially interact with previously created new role.
    # * TODO: Grant DROP role privileges to another user.
    # * TODO: Check role drop is available.

    # TODO: ================TABLEs interaction===============
    # ------------------READ---------------------------------
    # * READ is not available.
    with pytest.raises(
        TarantoolError,
        match=rf"Read access to space '{table_name}' is denied for user '{username}'",
    ):
        i1.sql(f""" select * from {table_name} """, user=username, password=password)
    # * Grant READ to user.
    acl = i1.sudo_sql(f""" grant read on table {table_name} to {username} """)
    assert acl["row_count"] == 1
    # * Granting already granted privilege do nothing.
    acl = i1.sql(f""" grant read on table {table_name} to {username} """)
    assert acl["row_count"] == 0
    # * After grant READ succeeds.
    i1.sql(f""" select * from {table_name} """, user=username, password=password)
    # * Revoke READ.
    acl = i1.sudo_sql(f""" revoke read on table {table_name} from {username} """)
    assert acl["row_count"] == 1
    # * After revoke READ fails again.
    with pytest.raises(
        TarantoolError,
        match=rf"Read access to space '{table_name}' is denied for user '{username}'",
    ):
        i1.sql(f""" select * from {table_name} """, user=username, password=password)
    # ------------------WRITE---------------------------------
    # TODO: remove
    acl = i1.sudo_sql(f""" grant read on table {table_name} to {username} """)
    assert acl["row_count"] == 1
    # * WRITE is not available.
    with pytest.raises(
        TarantoolError,
        match=rf"Write access to space '{table_name}' is denied for user '{username}'",
    ):
        i1.sql(
            f""" insert into {table_name} values (1) """,
            user=username,
            password=password,
        )
    # * Grant WRITE to user.
    acl = i1.sudo_sql(f""" grant write on table {table_name} to {username} """)
    assert acl["row_count"] == 1
    # * WRITE succeeds.
    i1.sql(
        f""" insert into {table_name} values (1) """, user=username, password=password
    )
    i1.sql(f""" delete from {table_name} where "a" = 1 """)
    # * Revoke WRITE from role.
    acl = i1.sudo_sql(f""" revoke write on table {table_name} from {username} """)
    assert acl["row_count"] == 1
    # * WRITE fails again.
    with pytest.raises(
        TarantoolError,
        match=rf"Write access to space '{table_name}' is denied for user '{username}'",
    ):
        i1.sql(
            f""" insert into {table_name} values (1) """,
            user=username,
            password=password,
        )
    # TODO: remove
    acl = i1.sudo_sql(f""" revoke read on table {table_name} from {username} """)
    assert acl["row_count"] == 1
    # ------------------CREATE---------------------------------
    # * TODO: Unable to create table.
    # * Grant CREATE to user.
    acl = i1.sudo_sql(f""" grant create table to {username} """)
    assert acl["row_count"] == 1
    # * TODO: Creation is available.
    # * TODO: Check user can do everything he wants on a table he created:
    # ** READ.
    # ** WRITE.
    # ** CREATE index.
    # ** ALTER index.
    # ** DROP.
    # * Revoke CREATE from user.
    acl = i1.sudo_sql(f""" revoke create table from {username} """)
    assert acl["row_count"] == 1
    # * TODO: Creation is not available again.
    # ------------------ALTER--------------------------------
    # * TODO: Unable to create new table index.
    # * Grant ALTER to user.
    acl = i1.sudo_sql(f""" grant alter on table {table_name} to {username} """)
    assert acl["row_count"] == 1
    # * TODO: Index creation succeeds.
    # * Revoke ALTER from user.
    acl = i1.sudo_sql(f""" revoke alter on table {table_name} from {username} """)
    assert acl["row_count"] == 1
    # * TODO: Attempt to remove index fails.
    # ------------------DROP---------------------------------
    # * TODO: Unable to drop table previously created by admin.
    # * Grant DROP to user.
    acl = i1.sudo_sql(f""" grant drop on table {table_name} to {username} """)
    assert acl["row_count"] == 1
    # * TODO: Able to drop admin table.
    # * Revoke DROP from user.
    acl = i1.sudo_sql(f""" revoke drop on table {table_name} from {username} """)
    assert acl["row_count"] == 1

    # Grant global tables READ, WRITE, ALTER, DROP.
    acl = i1.sudo_sql(f""" grant read table to {username} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" grant write table to {username} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" grant alter table to {username} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" grant drop table to {username} """)
    assert acl["row_count"] == 1
    # Check all operations available on another_table created by admin.
    i1.sql(
        f""" select * from {another_table_name} """, user=username, password=password
    )
    i1.sql(
        f""" insert into {another_table_name} values (1) """,
        user=username,
        password=password,
    )
    i1.sql(f""" delete from {another_table_name} """, user=username, password=password)
    i1.eval(
        f"""
        box.space.{another_table_name}:create_index('some', {{ parts = {{ 'a' }} }})
    """
    )
    i1.sql(f""" delete from {another_table_name} """, user=username, password=password)
    ddl = i1.sql(
        f""" drop table {another_table_name} """, user=username, password=password
    )
    assert ddl["row_count"] == 1
    # Revoke global privileges
    acl = i1.sudo_sql(f""" revoke read table from {username} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" revoke write table from {username} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" revoke alter table from {username} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" revoke drop table from {username} """)
    assert acl["row_count"] == 1

    # ================ROLE passing================
    # * Check there are no privileges granted to anything initially.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number
    # * Read from table is prohibited for user initially.
    with pytest.raises(
        TarantoolError,
        match=rf"Read access to space '{table_name}' is denied for user '{username}'",
    ):
        i1.sql(f""" select * from {table_name} """, user=username, password=password)
    # * Grant table READ and WRITE to role.
    acl = i1.sudo_sql(f""" grant read on table {table_name} to {rolename} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" grant write on table {table_name} to {rolename} """)
    assert acl["row_count"] == 1
    # * Grant ROLE to user.
    acl = i1.sudo_sql(f""" grant {rolename} to {username} """)
    assert acl["row_count"] == 1
    # * Check read and write is available for user.
    i1.sql(f""" select * from {table_name} """, user=username, password=password)
    i1.sql(
        f""" insert into {table_name} values (1) """, user=username, password=password
    )
    i1.sql(f""" delete from {table_name} where "a" = 1 """)
    # * Revoke privileges from role.
    acl = i1.sudo_sql(f""" revoke write on table {table_name} from {rolename} """)
    assert acl["row_count"] == 1
    acl = i1.sudo_sql(f""" revoke read on table {table_name} from {rolename} """)
    assert acl["row_count"] == 1
    # * Check privilege revoked from role and user.
    privs_rows = i1.sql(""" select * from "_pico_privilege" """)["rows"]
    assert len(privs_rows) == default_privileges_number + 1  # default + role for user
    # * Check read is prohibited again.
    with pytest.raises(
        TarantoolError,
        match=rf"Read access to space '{table_name}' is denied for user '{username}'",
    ):
        i1.sql(f""" select * from {table_name} """, user=username, password=password)


def test_distributed_sql_via_set_language(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, i2 = cluster.instances

    prelude = """
        local console = require('console')
        console.eval([[\\ set language sql]])
        console.eval([[\\ set delimiter ;]])
    """

    i1.eval(
        f"""
        {prelude}
        return console.eval('create table t \
            (a integer not null, b int not null, primary key (a)) \
                using memtx distributed by (b) option (timeout = 3);')
    """
    )

    i1.eval(
        f"""
        {prelude}
        return console.eval('insert into t values (22, 8);')
    """
    )

    select_from_second_instance = i2.eval(
        f"""
        {prelude}
        return console.eval('select * from t where a = 22;')
    """
    )

    assert (
        select_from_second_instance
        == """---
- metadata:
  - {'name': 'A', 'type': 'integer'}
  - {'name': 'B', 'type': 'integer'}
  rows:
  - [22, 8]
...
"""
    )


def test_sql_privileges(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    table_name = "t"
    # Create a test table
    ddl = i1.sql(
        f"""
        create table "{table_name}" ("a" int not null, "b" int, primary key ("a"))
        using memtx
        distributed by ("a")
        option (timeout = 3)
    """
    )
    assert ddl["row_count"] == 1

    username = "alice"
    alice_pwd = "Pa55sword"

    # Create user
    acl = i1.sql(
        f"""
        create user "{username}" with password '{alice_pwd}'
        using chap-sha1 option (timeout = 3)
    """
    )
    assert acl["row_count"] == 1

    # ------------------------
    # Check SQL read privilege
    # ------------------------
    with pytest.raises(TarantoolError, match=f"Read access to space '{table_name}'"):
        i1.sql(f""" select * from "{table_name}" """, user=username, password=alice_pwd)
    # Grant read privilege
    i1.sudo_sql(f""" grant read on table "{table_name}" to "{username}" """)
    dql = i1.sql(
        f""" select * from "{table_name}" """, user=username, password=alice_pwd
    )
    assert dql["rows"] == []

    # Revoke read privilege
    i1.sudo_sql(f""" revoke read on table "{table_name}" from "{username}" """)

    # -------------------------
    # Check SQL write privilege
    # -------------------------
    with pytest.raises(TarantoolError, match=f"Write access to space '{table_name}'"):
        i1.sql(
            f""" insert into "{table_name}" values (1, 2) """,
            user=username,
            password=alice_pwd,
        )

    # Grant write privilege
    i1.sudo_sql(f""" grant write on table "{table_name}" to "{username}" """)
    dml = i1.sql(
        f""" insert into "{table_name}" values (1, 2) """,
        user=username,
        password=alice_pwd,
    )
    assert dml["row_count"] == 1

    # Revoke write privilege
    i1.sudo_sql(f""" revoke write on table "{table_name}" from "{username}" """)

    # -----------------------------------
    # Check SQL write and read privileges
    # -----------------------------------
    with pytest.raises(TarantoolError, match=f"Read access to space '{table_name}'"):
        i1.sql(
            f""" insert into "{table_name}" select "a" + 1, "b" from "{table_name}"  """,
            user=username,
            password=alice_pwd,
        )
    with pytest.raises(TarantoolError, match=f"Read access to space '{table_name}'"):
        i1.sql(
            f""" update "{table_name}" set "b" = 42 """,
            user=username,
            password=alice_pwd,
        )
    with pytest.raises(TarantoolError, match=f"Read access to space '{table_name}'"):
        i1.sql(f""" delete from "{table_name}" """, user=username, password=alice_pwd)

    # Grant read privilege
    i1.sudo_sql(f""" grant read on table "{table_name}" to "{username}" """)

    with pytest.raises(TarantoolError, match=f"Write access to space '{table_name}'"):
        i1.sql(
            f""" insert into "{table_name}" select "a" + 1, "b" from "{table_name}"  """,
            user=username,
            password=alice_pwd,
        )
    with pytest.raises(TarantoolError, match=f"Write access to space '{table_name}'"):
        i1.sql(
            f""" update "{table_name}" set "b" = 42 """,
            user=username,
            password=alice_pwd,
        )
    with pytest.raises(TarantoolError, match=f"Write access to space '{table_name}'"):
        i1.sql(f""" delete from "{table_name}" """, user=username, password=alice_pwd)

    # Grant write privilege
    i1.sudo_sql(f""" grant write on table "{table_name}" to "{username}" """)

    dml = i1.sql(
        f""" insert into "{table_name}" select "a" + 1, "b" from "{table_name}"  """,
        user=username,
        password=alice_pwd,
    )
    assert dml["row_count"] == 1
    dml = i1.sql(
        f""" update "{table_name}" set "b" = 42 """, user=username, password=alice_pwd
    )
    assert dml["row_count"] == 2
    dml = i1.sql(f""" delete from "{table_name}" """, user=username, password=alice_pwd)
    assert dml["row_count"] == 2


def test_sql_privileges_vtables(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    username = "borat"
    pwd = "Password1"

    # Create user.
    acl = i1.sql(
        f"""
        create user "{username}" with password '{pwd}'
        using chap-sha1 option (timeout = 3)
    """
    )
    assert acl["row_count"] == 1

    # Grant create table privilege.
    i1.sudo_sql(f""" grant create table to "{username}" """)

    table_name = "foo"

    # Create a table by user.
    ddl = i1.sql(
        f"""
        create table "{table_name}" ("i" int not null, primary key ("i"))
        using memtx
        distributed by ("i")
    """,
        user=username,
        password=pwd,
    )
    assert ddl["row_count"] == 1

    # Check no-motion (without need to read vtables) query work.
    dql = i1.sql(f""" select * from "{table_name}" """, user=username, password=pwd)
    assert dql["rows"] == []

    # Check with-motion (with need to read vtables) query work.
    # See https://git.picodata.io/picodata/picodata/picodata/-/issues/620
    dql = i1.sql(
        f""" select count(*) from "{table_name}" """, user=username, password=pwd
    )
    assert dql["rows"] == [[0]]

    # Check DML query work.
    dml = i1.sql(
        f""" insert into "{table_name}" select count(*) from "{table_name}" """
    )
    assert dml["row_count"] == 1


def test_user_changes_password(cluster: Cluster):
    i1, *_ = cluster.deploy(instance_count=1)
    user_name = "U"
    old_password = "Passw0rd"
    new_password = "Pa55word"

    i1.create_user(with_name=user_name, with_password=old_password)

    i1.sql(
        f"""
        ALTER USER "{user_name}" PASSWORD '{new_password}'
        """,
        user=user_name,
        password=old_password,
    )
    # ensure we can authenticate with new password
    i1.sql("SELECT * FROM (VALUES (1))", user=user_name, password=new_password)


def test_create_drop_procedure(cluster: Cluster):
    i1, i2 = cluster.deploy(instance_count=2)

    data = i1.sql(
        """
        create table t (a int not null, b int, primary key (a))
        using memtx
        distributed by (b)
        """
    )
    assert data["row_count"] == 1

    # Check that the procedure would be created with the expected id.
    next_func_id = i1.eval("return box.internal.generate_func_id(true)")
    data = i2.sql(
        """
        create procedure proc1(int)
        language SQL
        as $$insert into t values(?, ?)$$
        """
    )
    assert data["row_count"] == 1
    data = i1.sql(
        """
        select "id" from "_pico_routine" where "name" = 'PROC1'
        """,
    )
    assert data["rows"] == [[next_func_id]]

    # Check that recreation of the same procedure is idempotent.
    data = i2.sql(
        """
        create procedure proc1(int)
        language SQL
        as $$insert into t values(?, ?)$$
        """
    )
    assert data["row_count"] == 0

    # Check that we can't create a procedure with the same name but different
    # signature.
    with pytest.raises(
        TarantoolError, match="routine PROC1 already exists with different parameters"
    ):
        i2.sql(
            """
            create procedure proc1(int, text)
            language SQL
            as $$insert into t values(?, ?)$$
            option(timeout=3)
            """
        )
    with pytest.raises(
        TarantoolError, match="routine PROC1 already exists with a different body"
    ):
        i2.sql(
            """
            create procedure proc1(int)
            language SQL
            as $$insert into t values(1, 2)$$
            """
        )

    # Check the routine creation abortion on _func space id conflict.
    i2.eval(
        """
        box.schema.func.create(
            'sum1',
            {
                body = [[function(a, b) return a + b end]],
                id = box.internal.generate_func_id(true)
            }
        )
        """
    )
    with pytest.raises(TarantoolError, match="ddl operation was aborted"):
        i1.sql(
            """
            create procedure proc2()
            as $$insert into t values(1, 2)$$
            """
        )
    data = i1.sql(""" select * from "_pico_routine" where "name" = 'PROC2' """)
    assert data["rows"] == []
    data = i2.sql(""" select * from "_pico_routine" where "name" = 'PROC2' """)
    assert data["rows"] == []

    # Check that PROC1 is actually dropped
    i1.sql(""" drop procedure proc1 """)
    data = i1.sql(""" select * from "_pico_routine" where "name" = 'PROC1' """)
    assert data["rows"] == []
    data = i2.sql(""" select * from "_pico_routine" where "name" = 'PROC1' """)
    assert data["rows"] == []

    # Check that dropping of the same procedure is idempotent.
    i1.sql(""" drop procedure proc1 """)
    i2.sql(""" drop procedure proc1 """)

    # Create proc for dropping.
    data = i2.sql(
        """
        create procedure proc1(int)
        language SQL
        as $$insert into t values(?, ?)$$
        """
    )
    assert data["row_count"] == 1

    data = i1.sql(
        """
        select "id" from "_pico_routine" where "name" = 'PROC1'
        """,
    )
    assert data["rows"] != []
    routine_id = data["rows"][0][0]

    # Check that dropping raises an error in case of parameters mismatch.
    with pytest.raises(
        TarantoolError,
        match=r"routine exists but with a different signature: PROC1\(integer\)",
    ):
        data = i2.sql(""" drop procedure proc1(decimal) """)

    # Check that dropping raises an error in case of parameters mismatch.
    with pytest.raises(
        TarantoolError,
        match=r"routine exists but with a different signature: PROC1\(integer\)",
    ):
        data = i2.sql(""" drop procedure proc1(integer, integer) """)

    # Routine mustn't be dropped at the moment.
    data = i1.sql(""" select * from "_pico_routine" where "name" = 'PROC1' """)
    assert data["rows"] != []
    data = i2.sql(""" select * from "_pico_routine" where "name" = 'PROC1' """)
    assert data["rows"] != []

    # Check drop with matching parameters.
    i2.sql(""" drop procedure proc1(integer) """)
    data = i1.sql(""" select * from "_pico_routine" where "name" = 'PROC1' """)
    assert data["rows"] == []
    data = i2.sql(""" select * from "_pico_routine" where "name" = 'PROC1' """)
    assert data["rows"] == []

    # Check that recreated routine has the same id with the recently dropped one.
    i2.sql(
        """
        create procedure proc1(int)
        language SQL
        as $$insert into t values(?, ?)$$
        """
    )
    data = i1.sql(
        """
        select "id" from "_pico_routine" where "name" = 'PROC1'
        """,
    )
    assert data["rows"] != []
    assert routine_id == data["rows"][0][0]


def test_sql_user_password_checks(cluster: Cluster):
    cluster.deploy(instance_count=2)
    i1, _ = cluster.instances

    with pytest.raises(
        TarantoolError,
        match="invalid password: password should contains at least one uppercase letter",
    ):
        i1.sql(
            """
            create user noname with password 'withoutdigitsanduppercase'
            using md5 option (timeout = 3)
            """
        )

    with pytest.raises(
        TarantoolError,
        match="invalid password: password should contains at least one lowercase letter",
    ):
        i1.sql(
            """
            create user noname with password 'PASSWORD3'
            using md5 option (timeout = 3)
            """
        )

    with pytest.raises(
        TarantoolError,
        match="invalid password: password should contains at least one digit",
    ):
        i1.sql(
            """
            create user noname with password 'Withoutdigits'
            using md5 option (timeout = 3)
            """
        )

    acl = i1.sql(
        """
        create user success with password 'Withdigit1'
        using md5 option (timeout = 3)
        """
    )
    assert acl["row_count"] == 1

    # let's turn off uppercase check and turn on special characters check
    read_index = i1.raft_read_index()

    ret = cluster.cas(
        "replace",
        "_pico_property",
        ["password_enforce_uppercase", False],
        index=read_index,
    )
    assert ret == read_index + 1

    ret = cluster.cas(
        "replace",
        "_pico_property",
        ["password_enforce_specialchars", True],
        index=read_index,
    )
    assert ret == read_index + 2

    cluster.raft_wait_index(ret)

    with pytest.raises(
        TarantoolError,
        match="invalid password: password should contains at least one special character",
    ):
        i1.sql(
            """
            create user noname with password 'withoutspecialchar14'
            using md5 option (timeout = 3)
            """
        )

    # now it's ok to create password without uppercase letter, but with special symbol
    acl = i1.sql(
        """
        create user noname with password '!withdigit1@'
        using md5 option (timeout = 3)
        """
    )
    assert acl["row_count"] == 1


def test_call_procedure(cluster: Cluster):
    i1, i2 = cluster.deploy(instance_count=2)

    data = i1.sql(
        """
        create table t (a int not null, b int, primary key (a))
        using memtx
        distributed by (b)
        """
    )
    assert data["row_count"] == 1

    # Check that the procedure is called successfully.
    data = i2.sql(
        """
        create procedure "proc1"(int, int)
        language SQL
        as $$insert into t values(?, ?) on conflict do nothing$$
        """
    )
    assert data["row_count"] == 1
    data = i2.sql(""" call "proc1"(1, 1) """)
    assert data["row_count"] == 1
    data = i2.sql(""" call "proc1"(1, 1) """)
    assert data["row_count"] == 0
    data = i2.sql(""" call "proc1"(2, 2) """)
    assert data["row_count"] == 1

    data = i1.sql(
        """
        create procedure "proc2"(int)
        language SQL
        as $$insert into t values(?, 42) on conflict do fail$$
        """
    )
    assert data["row_count"] == 1

    # Check that procedure returns an error when called with wrong number of arguments.
    with pytest.raises(TarantoolError, match=r"""expected 1 parameter\(s\), got 0"""):
        i2.sql(""" call "proc2"() """)
    with pytest.raises(TarantoolError, match=r"""expected 1 parameter\(s\), got 2"""):
        i2.sql(""" call "proc2"(3, 3) """)
    # Check that procedure returns an error when called with the wrong argument type.
    error_msg = "expected integer for parameter on position 0, got string"
    with pytest.raises(TarantoolError, match=error_msg):
        i2.sql(""" call "proc2"('hello') """)

    # Check internal statement errors are propagated.
    with pytest.raises(TarantoolError, match="Duplicate key exists in unique index"):
        i2.sql(""" call "proc2"(1) """)

    # Check parameters are passed correctly.
    data = i1.sql(""" call "proc2"(?) """, 4)
    assert data["row_count"] == 1
    data = i1.sql(
        """ call "proc2"($1) option(sql_vdbe_max_steps = $1, vtable_max_rows = $1)""",
        5,
    )
    assert data["row_count"] == 1

    # Check call permissions.
    username = "alice"
    alice_pwd = "Passw0rd"
    acl = i1.sql(
        f"""
        create user "{username}" with password '{alice_pwd}'
        using chap-sha1 option (timeout = 3)
    """
    )
    assert acl["row_count"] == 1
    with pytest.raises(
        TarantoolError, match="Execute access to function 'proc1' is denied"
    ):
        i1.sql(""" call "proc1"(3, 3) """, user=username, password=alice_pwd)


def test_rename_procedure(cluster: Cluster):
    i1, i2 = cluster.deploy(instance_count=2)

    data = i1.sql(
        """
        create table t (a int not null, b int, primary key (a))
        using memtx
        distributed by (b)
        """
    )
    assert data["row_count"] == 1

    data = i2.sql(
        """
        create procedure foo(int)
        language SQL
        as $$insert into t values(?, ?)$$
        """
    )
    assert data["row_count"] == 1

    data = i1.sql(
        """
        alter procedure foo
        rename to bar
        option ( timeout = 4 )
        """
    )
    assert data["row_count"] == 1

    data = i2.sql(
        """
        select "name" from "_pico_routine"
        where "name" = 'BAR' or "name" = 'FOO'
        """
    )
    assert data["rows"] == [["BAR"]]

    # procedure foo doesn't exist
    data = i1.sql(
        """
        alter procedure foo
        rename to bar
        """
    )
    assert data["row_count"] == 0

    # rename back, use syntax with parameters
    data = i1.sql(
        """
        alter procedure "BAR"(int)
        rename to "FOO"
        """
    )
    assert data["row_count"] == 1

    data = i2.sql(
        """
        select "name" from "_pico_routine"
        where "name" = 'BAR' or "name" = 'FOO'
        """
    )
    assert data["rows"] == [["FOO"]]

    data = i2.sql(
        """
        create procedure bar()
        language SQL
        as $$insert into t values(200, 200)$$
        """
    )
    assert data["row_count"] == 1

    with pytest.raises(TarantoolError, match="Name 'FOO' is already taken"):
        data = i1.sql(
            """
            alter procedure bar()
            rename to foo
            """
        )

    with pytest.raises(
        TarantoolError, match="routine exists but with a different signature: BAR()"
    ):
        data = i1.sql(
            """
            alter procedure bar(int)
            rename to buzz
            """
        )

    username = "alice"
    alice_pwd = "Passw0rd"
    acl = i1.sql(
        f"""
        create user "{username}" with password '{alice_pwd}'
        using chap-sha1 option (timeout = 3)
    """
    )
    assert acl["row_count"] == 1
    with pytest.raises(
        TarantoolError,
        match="Alter access to function 'BAR' is denied for user 'alice'",
    ):
        i1.sql(
            """alter procedure bar rename to buzz""", user=username, password=alice_pwd
        )

    i2.eval(
        """
        box.schema.func.create(
            'foobar',
            {
                body = [[function(question) return 42 end]],
                id = box.internal.generate_func_id(true)
            }
        )
        """
    )
    with pytest.raises(TarantoolError, match="ddl operation was aborted"):
        i1.sql(
            """
            alter procEdurE foo rENaME tO "foobar"
            """
        )

    data = i1.sql(""" select * from "_pico_routine" where "name" = 'foobar' """)
    assert data["rows"] == []
    data = i2.sql(""" select * from "_pico_routine" where "name" = 'foobar' """)
    assert data["rows"] == []


def test_procedure_privileges(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    table_name = "t"
    # Create a test table
    ddl = i1.sql(
        f"""
        create table "{table_name}" ("a" int not null, "b" int, primary key ("a"))
        using memtx
        distributed by ("a")
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    alice = "alice"
    alice_pwd = "Passw0rd"

    bob = "bob"
    bob_pwd = "Passw0rd"

    for user, pwd in [(alice, alice_pwd), (bob, bob_pwd)]:
        acl = i1.sql(
            f"""
            create user "{user}" with password '{pwd}'
            using chap-sha1 option (timeout = 3)
            """
        )
        assert acl["row_count"] == 1
        # grant write permission for procedure calls to work
        acl = i1.sudo_sql(
            f"""
            grant write on table "{table_name}" to "{user}"
            """
        )
        assert acl["row_count"] == 1

    # ---------------
    # HELPERS
    # ---------------

    def create_procedure(name: str, arg_cnt: int, as_user=None, as_pwd=None):
        assert arg_cnt < 3
        query = f"""
        create procedure {name}()
        language SQL
        as $$insert into "{table_name}" values(42, 8) on conflict do replace$$
        """
        if arg_cnt == 1:
            query = f"""
            create procedure {name}(int)
            language SQL
            as $$insert into "{table_name}" values($1, $1) on conflict do replace$$
            """
        if arg_cnt == 2:
            query = f"""
            create procedure {name}(int, int)
            language SQL
            as $$insert into "{table_name}" values($2, $1) on conflict do replace$$
            """
        ddl = None
        if not as_user:
            ddl = i1.sudo_sql(query)
        else:
            ddl = i1.sql(query, user=as_user, password=as_pwd)
        assert ddl["row_count"] == 1

    def drop_procedure(name: str, as_user=None, as_pwd=None):
        query = f"drop procedure {name}"
        ddl = None
        if not as_user:
            ddl = i1.sudo_sql(query)
        else:
            ddl = i1.sql(query, user=as_user, password=as_pwd)
        assert ddl["row_count"] == 1

    def rename_procedure(old_name: str, new_name: str, as_user=None, as_pwd=None):
        query = f"alter procedure {old_name} rename to {new_name}"
        ddl = None
        if not as_user:
            ddl = i1.sudo_sql(query)
        else:
            ddl = i1.sql(query, user=as_user, password=as_pwd)
        assert ddl["row_count"] == 1

    def grant_procedure(priv: str, user: str, fun=None, as_user=None, as_pwd=None):
        query = f"grant {priv}"
        if fun:
            query += f' on procedure "{fun}"'
        else:
            query += " procedure"
        query += f' to "{user}"'
        acl = (
            i1.sql(query, user=as_user, password=as_pwd)
            if as_user
            else i1.sudo_sql(query)
        )

        assert acl["row_count"] == 1

    def revoke_procedure(priv: str, user: str, fun=None, as_user=None, as_pwd=None):
        query = f"revoke {priv}"
        if fun:
            query += f' on procedure "{fun}"'
        else:
            query += " procedure"
        query += f' from "{user}"'
        acl = (
            i1.sql(query, user=as_user, password=as_pwd)
            if as_user
            else i1.sudo_sql(query)
        )

        assert acl["row_count"] == 1

    def call_procedure(proc, *args, as_user=None, as_pwd=None):
        args_str = ",".join(str(x) for x in args)
        data = i1.sql(
            f"""
            call {proc}({args_str})
            """,
            user=as_user,
            password=as_pwd,
        )
        assert data["row_count"] == 1

    def check_execute_access_denied(fun, username, pwd, *args):
        with pytest.raises(
            TarantoolError,
            match=f"Execute access to function '{fun}' "
            + f"is denied for user '{username}'",
        ):
            call_procedure(fun, *args, as_user=username, as_pwd=pwd)

    def check_create_access_denied(fun, username, pwd):
        with pytest.raises(
            TarantoolError,
            match=f"Create access to function '{fun}' "
            + f"is denied for user '{username}'",
        ):
            create_procedure(fun, 0, as_user=username, as_pwd=pwd)

    def check_drop_access_denied(fun, username, pwd):
        with pytest.raises(
            TarantoolError,
            match=f"Drop access to function '{fun}' "
            + f"is denied for user '{username}'",
        ):
            drop_procedure(fun, as_user=username, as_pwd=pwd)

    def check_rename_access_denied(old_name, new_name, username, pwd):
        with pytest.raises(
            TarantoolError,
            match=f"Alter access to function '{old_name}' "
            + f"is denied for user '{username}'",
        ):
            rename_procedure(old_name, new_name, as_user=username, as_pwd=pwd)

    # ----------------- Default privliges -----------------

    # Check that a user can't create a procedure without permition.
    check_create_access_denied("FOOBAZSPAM", alice, alice_pwd)

    # Check that a non-owner user without drop privilege can't drop proc
    create_procedure("FOO", 0)
    check_drop_access_denied("FOO", bob, bob_pwd)

    # Check that a non-owner user can't rename proc
    check_rename_access_denied("FOO", "BAR", bob, bob_pwd)

    # Check that a user without permession can't call proc
    check_execute_access_denied("FOO", bob, bob_pwd)

    # Check that owner can call proc
    call_procedure("FOO")

    # Check that owner can rename proc
    rename_procedure("FOO", "BAR")

    # Check that owner can drop proc
    drop_procedure("BAR")

    # ----------------- Default privliges -----------------

    # ----------------- grant-revoke privilege -----------------

    # ALL PROCEDURES

    # Check admin can grant create procedure to user
    grant_procedure("create", alice)
    create_procedure("FOO", 0, alice, alice_pwd)
    drop_procedure("FOO", alice, alice_pwd)
    create_procedure("FOO", 0)
    check_drop_access_denied("FOO", alice, alice_pwd)
    check_rename_access_denied("FOO", "BAR", alice, alice_pwd)
    drop_procedure("FOO")

    # Check admin can revoke create procedure from user
    revoke_procedure("create", alice)
    check_create_access_denied("FOO", alice, alice_pwd)

    # check grant execute to all procedures
    grant_procedure("create", bob)
    create_procedure("FOO", 0, bob, bob_pwd)
    grant_procedure("execute", alice)
    call_procedure("FOO", as_user=alice, as_pwd=alice_pwd)

    # check revoke execute from all procedures
    revoke_procedure("execute", alice)
    check_execute_access_denied("FOO", alice, alice_pwd)
    revoke_procedure("create", bob)

    # check grant drop for all procedures
    drop_procedure("FOO")
    create_procedure("FOO", 0)
    check_drop_access_denied("FOO", bob, bob_pwd)
    grant_procedure("drop", bob)
    drop_procedure("FOO", bob, bob_pwd)

    # check revoke drop for all procedures
    create_procedure("FOO", 0)
    revoke_procedure("drop", bob)
    check_drop_access_denied("FOO", bob, bob_pwd)

    drop_procedure("FOO")

    # Check that user can't grant create procedure (only admin can)
    with pytest.raises(
        TarantoolError,
        match=f"Grant to routine '' is denied for user '{alice}'",
    ):
        grant_procedure("create", bob, as_user=alice, as_pwd=alice_pwd)

    # Check that user can't grant execute procedure (only admin can)
    with pytest.raises(
        TarantoolError,
        match=f"Grant to routine '' is denied for user '{alice}'",
    ):
        grant_procedure("execute", bob, as_user=alice, as_pwd=alice_pwd)

    # SPECIFIC PROCEDURE

    # check grant execute specific procedure
    grant_procedure("create", alice)
    create_procedure("FOO", 0, alice, alice_pwd)
    check_execute_access_denied("FOO", bob, bob_pwd)
    grant_procedure("execute", bob, "FOO", as_user=alice, as_pwd=alice_pwd)
    call_procedure("FOO", as_user=bob, as_pwd=bob_pwd)

    # check admin can revoke execute from user
    revoke_procedure("execute", bob, "FOO", alice, alice_pwd)
    check_execute_access_denied("FOO", bob, bob_pwd)

    # check owner of procedure can grant drop to other user
    check_drop_access_denied("FOO", bob, bob_pwd)
    grant_procedure("drop", bob, "FOO", as_user=alice, as_pwd=alice_pwd)
    check_rename_access_denied("FOO", "BAR", bob, bob_pwd)
    drop_procedure("FOO", bob, bob_pwd)

    # check owner of procedure can revoke drop to other user
    create_procedure("FOO", 0, alice, alice_pwd)
    check_drop_access_denied("FOO", bob, bob_pwd)
    grant_procedure("drop", bob, "FOO", as_user=alice, as_pwd=alice_pwd)
    revoke_procedure("drop", bob, "FOO", alice, alice_pwd)
    check_drop_access_denied("FOO", bob, bob_pwd)

    # check we can't grant create specific procedure
    with pytest.raises(
        TarantoolError,
        match="sbroad: invalid privilege",
    ):
        grant_procedure("create", bob, "FOO")

    # check we can't grant to non-existing user
    with pytest.raises(
        TarantoolError,
        match="Nor user, neither role with name pasha exists",
    ):
        grant_procedure("drop", "pasha", "FOO")

    # check we can't revoke from non-existing user
    with pytest.raises(
        TarantoolError,
        match="Nor user, neither role with name pasha exists",
    ):
        revoke_procedure("execute", "pasha", "FOO")


def test_rename_user(cluster: Cluster):
    i1, _ = cluster.deploy(instance_count=2)

    # Two users biba and boba
    biba = "biba"
    boba = "boba"
    password = "Passw0rd"

    i1.create_user(with_name=biba, with_password=password)
    i1.create_user(with_name=boba, with_password=password)

    # rename to it's own name, nothing happens
    data = i1.sql(
        f"""
        ALTER USER "{boba}"
        RENAME TO "{boba}"
        """,
        user=boba,
        password=password,
    )
    assert data["row_count"] == 0

    # Existed name
    with pytest.raises(
        TarantoolError,
        match=f'User with name "{biba}" exists. Unable to rename user "{boba}"',
    ):
        data = i1.sudo_sql(
            f"""
            ALTER USER "{boba}"
            RENAME TO "{biba}"
            """
        )

    # Without privileges they cannot rename each other
    with pytest.raises(
        TarantoolError,
        match="""\
Alter access to user 'boba' is denied for user 'biba'\
""",
    ):
        data = i1.sql(
            f"""
            ALTER USER "{boba}"
            RENAME TO "skibidi"
            """,
            user=biba,
            password=password,
        )

    # Admin can rename anyone: boba -> skibidi
    data = i1.sudo_sql(
        f"""
        ALTER USER "{boba}"
        RENAME TO "skibidi"
        """
    )
    assert data["row_count"] == 1

    def names_from_pico_user_table():
        data = i1.sudo_sql(
            """
            SELECT "name" FROM "_pico_user"
            """
        )

        return [row[0] for row in data["rows"]]

    # Check that rename works fine
    assert "skibidi" in names_from_pico_user_table()
    assert boba not in names_from_pico_user_table()

    # Check that we can create a new user "boba" after rename him
    i1.create_user(with_name=boba, with_password=password)
    assert boba in names_from_pico_user_table()


def test_index(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    # Sharded memtx table
    ddl = i1.sql(
        """
        create table t (a text not null, b text not null, c text, primary key (a))
        using memtx
        distributed by (a)
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    # Global table
    ddl = i1.sql(
        """
        create table g (a int not null, b text not null, c text, primary key (a))
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1

    # Sharded vinyl table
    ddl = i1.sql(
        """
        create table v (a int not null, b text not null, c text, primary key (a))
        using vinyl
        distributed by (a)
        option (timeout = 3)
        """
    )

    # Check that created index appears in _pico_index table.
    ddl = i1.sql(""" create index i0 on t (a) option (timeout = 3) """)
    assert ddl["row_count"] == 1
    data = i1.sql(""" select * from "_pico_index" where "name" = 'I0' """)
    assert data["rows"] != []

    # Successful tree index creation with default options
    ddl = i1.sql(""" create index i1 on t (a, b) """)
    assert ddl["row_count"] == 1

    # Unique index can be created only on the sharding key for sharded tables.
    invalid_unique = (
        "unique index for the sharded table must duplicate its sharding key columns"
    )
    with pytest.raises(TarantoolError, match=invalid_unique):
        i1.sql(""" create unique index i2 on t using tree (b) """)

    # Successful unique tree index creation on the sharding key.
    ddl = i1.sql(""" create unique index i2 on t using tree (a) """)
    assert ddl["row_count"] == 1

    # No restrictions on unique index for globally distributed tables.
    ddl = i1.sql(""" create unique index i3 on g using tree (b) """)

    # Successful create a tree index with corresponding options.
    ddl = i1.sql(""" create index i4 on t using tree (c) with (HiNT = TrUe) """)
    assert ddl["row_count"] == 1

    # Fail to create a tree index with wrong options.
    invalid_tree_option = "index type tree does not support option"
    with pytest.raises(TarantoolError, match=invalid_tree_option):
        i1.sql(""" create index i5 on t using tree (c) with (distance = euclid) """)
    with pytest.raises(TarantoolError, match=invalid_tree_option):
        i1.sql(""" create index i6 on t using tree (c) with (dimension = 42) """)

    # RTree indexes can't be created via SQL at the moment as they require array columns
    # that are not supported yet.
    non_array_rtree = "index type rtree does not support column type"
    with pytest.raises(TarantoolError, match=non_array_rtree):
        i1.sql(""" create index i11 on t using rtree (b) """)

    # Fail to create an rtree index from nullable columns.
    nullable_rtree = "index type rtree does not support nullable columns"
    with pytest.raises(TarantoolError, match=nullable_rtree):
        i1.sql(""" create index i12 on t using rtree (c) """)

    # Fail to create an rtree index with wrong options.
    invalid_rtree_option = "index type rtree does not support option"
    with pytest.raises(TarantoolError, match=invalid_rtree_option):
        i1.sql(""" create index i13 on t using rtree (b) with (hint = true) """)

    # Successful bitset index creation.
    ddl = i1.sql(""" create index i14 on t using bitset (b) """)

    # Fail to create a bitset index from nullable columns.
    nullable_bitset = "index type bitset does not support nullable columns"
    with pytest.raises(TarantoolError, match=nullable_bitset):
        i1.sql(""" create index i15 on t using bitset (c) """)

    # Fail to create unique bitset index.
    unique_bitset = "index type bitset does not support unique indexes"
    with pytest.raises(TarantoolError, match=unique_bitset):
        i1.sql(""" create unique index i16 on t using bitset (a) """)

    # Fail to create bitset index with column types other then string, number or varbinary.
    invalid_bitset = "index type bitset does not support column type"
    with pytest.raises(TarantoolError, match=invalid_bitset):
        i1.sql(""" create index i17 on v using bitset (a) """)

    # Successful hash index creation.
    ddl = i1.sql(""" create unique index i17 on t using hash (a) """)
    assert ddl["row_count"] == 1

    # Fail to create a non-unique hash index.
    non_unique_hash = "index type hash does not support non-unique indexes"
    with pytest.raises(TarantoolError, match=non_unique_hash):
        i1.sql(""" create index i18 on t using hash (c) """)

    # Fail to create an index on memtex table with vinyl options.
    invalid_memtx = "table engine memtx does not support option"
    with pytest.raises(TarantoolError, match=invalid_memtx):
        i1.sql(""" create index i7 on t (b) with (page_size = 42) """)
    with pytest.raises(TarantoolError, match=invalid_memtx):
        i1.sql(""" create index i8 on t (b) with (range_size = 42) """)
    with pytest.raises(TarantoolError, match=invalid_memtx):
        i1.sql(""" create index i9 on t (b) with (run_count_per_level = 42) """)
    with pytest.raises(TarantoolError, match=invalid_memtx):
        i1.sql(""" create index i10 on t (b) with (run_size_ratio = 0.1) """)

    # Successful index drop.
    ddl = i1.sql(""" drop index i0 """)
    assert ddl["row_count"] == 1

    # Check that the index is actually dropped.
    data = i1.sql(""" select * from "_pico_index" where "name" = 'I0' """)
    assert data["rows"] == []

    # Drop non-existing index.
    ddl = i1.sql(""" drop index i0 option (timeout = 3) """)
    assert ddl["row_count"] == 0

    ddl = i1.sql(""" create index i19 on t (b)""")
    assert ddl["row_count"] == 1
    with pytest.raises(TarantoolError, match="index I19 already exists"):
        i1.sql(""" create index i19 on v (b)""")


def test_order_by(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    ddl = i1.sql(
        """
        create table "null_t" ("na" integer not null, "nb" int, "nc" int, primary key ("na"))
        distributed by ("na")
    """
    )
    assert ddl["row_count"] == 1

    dml = i1.sql(
        """
        insert into "null_t" values
            (1, 2, 1),
            (2, NULL, 3),
            (3, 2, 3),
            (4, 3, 1),
            (5, 1, 5),
            (6, -1, 3),
            (7, 1, 1),
            (8, NULL, -1)
    """
    )
    assert dml["row_count"] == 8

    expected_ordering_by_1 = [
        [1, 2, 1],
        [2, None, 3],
        [3, 2, 3],
        [4, 3, 1],
        [5, 1, 5],
        [6, -1, 3],
        [7, 1, 1],
        [8, None, -1],
    ]
    data = i1.sql(""" select * from "null_t" order by "na" """)
    assert data["rows"] == expected_ordering_by_1
    data = i1.sql(""" select * from "null_t" order by 1 """)
    assert data["rows"] == expected_ordering_by_1
    data = i1.sql(""" select * from "null_t" order by 1 asc """)
    assert data["rows"] == expected_ordering_by_1
    data = i1.sql(""" select * from "null_t" order by "na" asc """)
    assert data["rows"] == expected_ordering_by_1
    data = i1.sql(""" select * from "null_t" order by 1, 2 """)
    assert data["rows"] == expected_ordering_by_1

    expected_ordering_by_2 = [
        [2, None, 3],
        [8, None, -1],
        [6, -1, 3],
        [5, 1, 5],
        [7, 1, 1],
        [1, 2, 1],
        [3, 2, 3],
        [4, 3, 1],
    ]
    data = i1.sql(""" select * from "null_t" order by "nb" """)
    assert data["rows"] == expected_ordering_by_2
    data = i1.sql(""" select * from "null_t" order by "nb" asc """)
    assert data["rows"] == expected_ordering_by_2
    data = i1.sql(""" select * from "null_t" order by 2 """)
    assert data["rows"] == expected_ordering_by_2

    data = i1.sql(""" select * from "null_t" order by 1 desc """)
    assert data["rows"] == [
        [8, None, -1],
        [7, 1, 1],
        [6, -1, 3],
        [5, 1, 5],
        [4, 3, 1],
        [3, 2, 3],
        [2, None, 3],
        [1, 2, 1],
    ]

    expected_ordering_by_2_desc = [
        [4, 3, 1],
        [1, 2, 1],
        [3, 2, 3],
        [5, 1, 5],
        [7, 1, 1],
        [6, -1, 3],
        [2, None, 3],
        [8, None, -1],
    ]
    data = i1.sql(""" select * from "null_t" order by "nb" desc """)
    assert data["rows"] == expected_ordering_by_2_desc
    data = i1.sql(""" select * from "null_t" order by "nb" * 2 + 42 * "nb" desc """)
    assert data["rows"] == expected_ordering_by_2_desc

    data = i1.sql(""" select * from "null_t" order by "nb" desc, "na" desc """)
    assert data["rows"] == [
        [4, 3, 1],
        [3, 2, 3],
        [1, 2, 1],
        [7, 1, 1],
        [5, 1, 5],
        [6, -1, 3],
        [8, None, -1],
        [2, None, 3],
    ]

    data = i1.sql(""" select * from "null_t" order by 2 asc, 1 desc, 2 desc, 1 asc """)
    assert data["rows"] == [
        [8, None, -1],
        [2, None, 3],
        [6, -1, 3],
        [7, 1, 1],
        [5, 1, 5],
        [3, 2, 3],
        [1, 2, 1],
        [4, 3, 1],
    ]

    with pytest.raises(
        TarantoolError,
        match="Ordering index \\(4\\) is bigger than child projection output length \\(3\\)",
    ):
        i1.sql(""" select * from "null_t" order by 4 """)
    with pytest.raises(
        TarantoolError,
        match="Using parameter as a standalone ORDER BY expression doesn't influence sorting",
    ):
        i1.sql(""" select * from "null_t" order by ? """)


def test_cte(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]

    # Initialize sharded table
    ddl = i1.sql(
        """
        create table t (a int not null, b int, primary key (a))
        distributed by (b)
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1
    data = i1.sql(
        """
        insert into t values (1, 1), (2, 2), (3, 3), (4, 4), (5, 5)
        """
    )
    assert data["row_count"] == 5

    # Initialize global table
    ddl = i1.sql(
        """
        create table g (a int not null, b int, primary key (a))
        distributed globally
        option (timeout = 3)
        """
    )
    assert ddl["row_count"] == 1
    for i in range(1, 5):
        index = i1.cas("insert", "G", [i, i])
        i1.raft_wait_index(index, 3)

    # basic CTE
    data = i1.sql(
        """
        with cte (b) as (select a from t where a > 2)
        select * from cte
        """
    )
    assert data["rows"] == [[3], [4], [5]]

    # nested CTE
    data = i1.sql(
        """
        with cte1 (b) as (select a from t where a > 2),
             cte2 as (select b from cte1)
        select * from cte2
        """
    )
    assert data["rows"] == [[3], [4], [5]]

    # reuse CTE
    data = i1.sql(
        """
        with cte (b) as (select a from t where a = 2)
        select b from cte
        union all
        select b + b from cte
        """
    )
    assert data["rows"] == [[2], [4]]

    # global CTE
    data = i1.sql(
        """
        with cte (b) as (select a from g where a < 4)
        select * from cte
        """
    )
    assert data["rows"] == [[1], [2], [3]]

    # CTE with parameters
    data = i1.sql(
        """
        with cte (b) as (select a from t where a > ?)
        select * from cte
        """,
        3,
    )
    assert data["rows"] == [[4], [5]]

    # join sharded table with CTE
    data = i1.sql(
        """
        with cte (b) as (select a from t where a > 2)
        select t.a, cte.b from t join cte on t.a = cte.b
        """
    )
    assert data["rows"] == [[3, 3], [4, 4], [5, 5]]

    # join global table with CTE
    data = i1.sql(
        """
        with cte (b) as (select a from g where a < 4)
        select g.a, cte.b from g join cte on g.a = cte.b
        """
    )
    assert data["rows"] == [[1, 1], [2, 2], [3, 3]]

    # CTE in aggregate
    data = i1.sql(
        """
        with cte (b) as (select a from t where a between 2 and 4)
        select count(b) from cte
        """
    )
    assert data["rows"] == [[3]]

    # CTE in subquery
    data = i1.sql(
        """
        with cte (b) as (select a from t where a in (2, 5))
        select * from t where a in (select b from cte)
        """
    )
    assert data["rows"] == [[2, 2], [5, 5]]

    # values in CTE
    data = i1.sql(
        """
        with cte (b) as (values (1), (2), (3))
        select * from cte order by 1
        """
    )
    assert data["rows"] == [[1], [2], [3]]

    # union all in CTE
    data = i1.sql(
        """
        with cte1 (b) as (values (1), (2), (3)),
        cte2 (b) as (select a from t where a = 1 union all select * from cte1)
        select * from cte2 order by 1
        """
    )
    assert data["rows"] == [[1], [1], [2], [3]]

    # join in CTE
    data = i1.sql(
        """
        with cte (c) as (
            select t.a from t join g on t.a = g.a
            join t as t2 on t2.a = t.a
            where t.a = 1
        )
        select * from cte
        """
    )
    assert data["rows"] == [[1]]

    # order by in CTE
    data = i1.sql(
        """
        with cte (b) as (select a from t order by a desc)
        select * from cte
        """
    )
    assert data["rows"] == [[5], [4], [3], [2], [1]]

    # randomly distributed CTE used multiple times
    data = i1.sql(
        """
        with cte (b) as (select a from t where a > 3)
        select t.c from (select count(*) as c from cte c1 join cte c2 on true) t
        join cte on true
        """
    )
    assert data["rows"] == [[4], [4]]

    # CTE with segment distribution used multiple times
    data = i1.sql(
        """
        with cte (b) as (select a from t where a = 1)
        select t.c from (select count(*) as c from cte c1 join cte c2 on true) t
        join cte on true
        """
    )
    assert data["rows"] == [[1]]

    # CTE with global distribution used multiple times
    data = i1.sql(
        """
        with cte (b) as (select a from g where a = 1)
        select t.c from (select count(*) as c from cte c1 join cte c2 on true) t
        join cte on true
        """
    )
    assert data["rows"] == [[1]]

    # CTE with values used multiple times
    data = i1.sql(
        """
        with cte (b) as (values (1))
        select t.c from (select count(*) as c from cte c1 join cte c2 on true) t
        join cte on true
        """
    )
    assert data["rows"] == [[1]]


def test_unique_index_name_for_sharded_table(cluster: Cluster):
    cluster.deploy(instance_count=1)
    i1 = cluster.instances[0]
    table_names = ["T", "T2"]

    # Initialize two sharded table
    for table_name in table_names:
        ddl = i1.sql(
            f"""
            create table {table_name} (a int not null, b int, primary key (a))
            distributed by (b)
            option (timeout = 3)
            """
        )
        assert ddl["row_count"] == 1

    for table_name, other_table_name in zip(table_names, reversed(table_names)):
        with pytest.raises(
            TarantoolError,
            match=f"""index {table_name}_bucket_id already exists with a different signature""",
        ):
            # try to create existing index
            i1.sql(
                f""" create index "{table_name}_bucket_id"
                on "{table_name}" (a) option (timeout = 3) """
            )

        with pytest.raises(
            TarantoolError,
            match=f"""index {other_table_name}_bucket_id already exists""",
        ):
            # try to create non existing index with existing name
            i1.sql(
                f""" create index "{other_table_name}_bucket_id"
                on "{table_name}" (a) option (timeout = 3) """
            )

        # ensure that index on field bucket_id of sharded table exists
        data = i1.sql(
            f""" select * from "_pico_index" where "name" = '{table_name}_bucket_id' """
        )
        assert data["rows"] != []
